# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import glob
import os
import re
import signal
import subprocess
from enum import Enum
from lib389._constants import TaskWarning
from lib389.cli_base import CustomHelpFormatter
from lib389.dseldif import DSEldif
from pathlib import Path


class IndexOrdering(Enum):
    """Represents the ordering type of an index."""
    INTEGER = "integer"
    LEXICOGRAPHIC = "lexicographic"
    UNKNOWN = "unknown"


def dbtasks_db2index(inst, log, args):
    inst.log = log
    if not inst.db2index(bename=args.backend, attrs=args.attr):
        log.fatal("db2index failed")
        return False
    else:
        log.info("db2index successful")
        return True


def dbtasks_db2bak(inst, log, args):
    # Needs an output name?
    inst.log = log
    if not inst.db2bak(args.archive, watch=args.watch):
        log.fatal("db2bak failed")
        return False
    else:
        log.info("db2bak successful")


def dbtasks_bak2db(inst, log, args):
    # Needs the archive to restore.
    inst.log = log
    if not inst.bak2db(args.archive, watch=args.watch):
        log.fatal("bak2db failed")
        return False
    else:
        log.info("bak2db successful")


def dbtasks_db2ldif(inst, log, args):
    # If export filename is provided, check if file path exists
    inst.log = log
    if args.ldif:
        path = Path(args.ldif)
        parent = path.parent.absolute()
        if not parent.exists():
            raise ValueError("The LDIF export location does not exist: "
                             + args.ldif)

    # Export backend
    if not inst.db2ldif(bename=args.backend, encrypt=args.encrypted, repl_data=args.replication,
                        outputfile=args.ldif, suffixes=None, excludeSuffixes=None, export_cl=False,
                        watch=args.watch):
        log.fatal("db2ldif failed")
        return False
    else:
        log.info("db2ldif successful")


def dbtasks_ldif2db(inst, log, args):
    # Check if ldif file exists
    inst.log = log
    if not os.path.exists(args.ldif):
        raise ValueError("The LDIF file does not exist: " + args.ldif)

    ret = inst.ldif2db(bename=args.backend, encrypt=args.encrypted, import_file=args.ldif,
                       suffixes=None, excludeSuffixes=None, import_cl=False, watch=args.watch)
    if not ret:
        log.fatal("ldif2db failed")
        return False
    elif ret == TaskWarning.WARN_SKIPPED_IMPORT_ENTRY:
        log.warn("ldif2db successful with skipped entries")
    else:
        log.info("ldif2db successful")


def dbtasks_backups(inst, log, args):
    inst.log = log
    if args.delete:
        # Delete backup
        inst.del_backup(args.delete[0])
    else:
        # list backups
        if not inst.backups(args.json):
            log.fatal("Failed to get list of backups")
            return False
        else:
            if args.json is None:
                log.info("backups successful")


def dbtasks_ldifs(inst, log, args):
    inst.log = log
    if args.delete:
        # Delete LDIF file
        inst.del_ldif(args.delete[0])
    else:
        # list LDIF files
        if not inst.ldifs(args.json):
            log.fatal("Failed to get list of LDIF files")
            return False
        else:
            if args.json is None:
                log.info("backups successful")


def dbtasks_verify(inst, log, args):
    inst.log = log
    if not inst.dbverify(bename=args.backend):
        log.fatal("dbverify failed")
        return False
    else:
        log.info("dbverify successful")


def _get_db_dir(dse_ldif):
    """Get the database directory.

    Args:
        dse_ldif: DSEldif instance.

    Returns:
        Path to the database directory, or None if not found.
    """
    try:
        db_dir = dse_ldif.get(
            "cn=config,cn=ldbm database,cn=plugins,cn=config",
            "nsslapd-directory",
            single=True,
        )
        return db_dir
    except (ValueError, TypeError):
        pass
    return None



def _has_integer_ordering_match(dse_ldif, backend, index_name):
    """Check if an index has integerOrderingMatch configured in DSE.

    Args:
        dse_ldif: DSEldif instance.
        backend: Backend name.
        index_name: Name of the index to check.

    Returns:
        True if integerOrderingMatch is configured, False otherwise.
    """
    index_dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
        index_name, backend
    )
    matching_rules = dse_ldif.get(index_dn, "nsMatchingRule", lower=True)
    if matching_rules:
        return any(mr.lower() == "integerorderingmatch" for mr in matching_rules)
    return False


def _has_index_scan_limit(dse_ldif, backend, index_name):
    """Check if an index has nsIndexIDListScanLimit configured.

    Args:
        dse_ldif: DSEldif instance.
        backend: Backend name.
        index_name: Name of the index to check.

    Returns:
        True if nsIndexIDListScanLimit is configured, False otherwise.
    """
    index_dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
        index_name, backend
    )
    scan_limit = dse_ldif.get(index_dn, "nsIndexIDListScanLimit")
    return scan_limit is not None


def _index_config_exists(dse_ldif, backend, index_name):
    """Check if an index configuration entry exists in DSE.

    Args:
        dse_ldif: DSEldif instance.
        backend: Backend name.
        index_name: Name of the index to check.

    Returns:
        True if the index config entry exists, False otherwise.
    """
    index_dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
        index_name, backend
    )
    try:
        cn = dse_ldif.get(index_dn, "cn")
        return cn is not None
    except (ValueError, KeyError):
        return False


def _default_index_exists(dse_ldif, index_name):
    """Check if an index exists in cn=default indexes.

    Args:
        dse_ldif: DSEldif instance.
        index_name: Name of the index to check.

    Returns:
        True if the index exists in default indexes, False otherwise.
    """
    index_dn = "cn={},cn=default indexes,cn=config,cn=ldbm database,cn=plugins,cn=config".format(
        index_name
    )
    try:
        cn = dse_ldif.get(index_dn, "cn")
        return cn is not None
    except (ValueError, KeyError):
        return False


def _check_disk_ordering(db_dir, backend, index_name, dbscan_path, is_mdb, log):
    """Check if index on disk uses lexicographic or integer ordering.

    Args:
        db_dir: Path to the database directory.
        backend: Backend name.
        index_name: Name of the index to check.
        dbscan_path: Path to the dbscan binary.
        is_mdb: True if using MDB backend.
        log: Logger instance.

    Returns:
        IndexOrdering: The detected ordering type.
    """
    if is_mdb:
        # MDB uses pseudo-paths: db_dir/backend/index.db
        # dbscan accesses indexes via paths like: /var/lib/dirsrv/slapd-xxx/db/userroot/parentid.db
        index_file = os.path.join(db_dir, backend, "{}.db".format(index_name))
    else:
        # BDB has separate directories per backend with actual index files
        backend_dir = os.path.join(db_dir, backend)
        if not os.path.exists(backend_dir):
            return IndexOrdering.UNKNOWN
        index_file = None
        pattern = os.path.join(backend_dir, "{}.db*".format(index_name))
        for f in glob.glob(pattern):
            if os.path.isfile(f):
                index_file = f
                break
        if not index_file:
            return IndexOrdering.UNKNOWN

    # Only read the first 100 lines from dbscan to avoid scanning the
    # entire index (which can take hours on large databases).
    try:
        proc = subprocess.Popen(
            [dbscan_path, "-f", index_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        keys = []
        line_count = 0
        assert proc.stdout is not None
        for line in proc.stdout:
            line_count += 1
            if line_count > 100:
                break
            line = line.strip()
            if line.startswith("="):
                match = re.match(r"^=(\d+)", line)
                if match:
                    keys.append(int(match.group(1)))

        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()
            proc.wait()

        if proc.returncode not in (0, -signal.SIGTERM):
            log.warning("  dbscan returned non-zero exit code for %s", index_file)
            return IndexOrdering.UNKNOWN

        if len(keys) < 2:
            return IndexOrdering.UNKNOWN

        # Check if keys are in integer order by looking for decreasing numeric values
        # (which would indicate lexicographic ordering, e.g., "3" < "30" < "4")
        prev_id = keys[0]
        for current_id in keys[1:]:
            if prev_id > current_id:
                return IndexOrdering.LEXICOGRAPHIC
            prev_id = current_id

        return IndexOrdering.INTEGER

    except OSError as e:
        log.warning("  Error running dbscan: %s", e)
        return IndexOrdering.UNKNOWN


def dbtasks_index_check(inst, log, args):
    """Check and optionally fix index ordering mismatches.

    This function detects mismatches between the configured ordering
    (integerOrderingMatch in DSE) and the actual on-disk ordering of
    parentid and ancestorid indexes.

    Args:
        inst: DirSrv instance.
        log: Logger instance.
        args: Parsed command line arguments.

    Returns:
        True if all checks passed, False if mismatches were detected.
    """
    # Server must be stopped
    if inst.status():
        log.error("index-check requires the instance to be stopped")
        return False

    # Check for dbscan binary
    dbscan_path = os.path.join(inst.ds_paths.bin_dir, "dbscan")
    if not os.path.exists(dbscan_path):
        log.error("dbscan utility not found at %s", dbscan_path)
        return False

    # Load DSE
    try:
        dse_ldif = DSEldif(inst)
    except Exception as e:
        log.error("Failed to read dse.ldif: %s", e)
        return False

    # Get backends to check
    all_backends = dse_ldif.get_backends()
    if not all_backends:
        log.info("No backends found")
        return True

    # Filter to specific backend if requested
    if args.backend:
        # Case-insensitive backend lookup
        backend_lower = args.backend.lower()
        matching_backend = None
        for be in all_backends:
            if be.lower() == backend_lower:
                matching_backend = be
                break
        if matching_backend is None:
            log.error("Backend '%s' not found. Available backends: %s",
                      args.backend, ", ".join(all_backends))
            return False
        backends_to_check = [matching_backend]
    else:
        backends_to_check = all_backends

    # Get database directory and check database type
    db_dir = _get_db_dir(dse_ldif)
    if not db_dir or not os.path.exists(db_dir):
        log.error("Database directory not found")
        return False

    db_lib = inst.get_db_lib()
    is_mdb = (db_lib == "mdb")
    log.info("Database type: %s", db_lib.upper())

    # Track all issues found
    all_ok = True
    config_fixes = []  # (backend, index_name, action) tuples: action is "add_mr" or "remove_mr"
    scan_limits_to_remove = []  # (backend, index_name) tuples with nsIndexIDListScanLimit
    ancestorid_configs_to_remove = []  # backend names with ancestorid config entries
    remove_ancestorid_from_defaults = False  # Flag to remove from cn=default indexes

    # Check if ancestorid exists in cn=default indexes (should be removed)
    if _default_index_exists(dse_ldif, "ancestorid"):
        log.warning("ancestorid found in cn=default indexes - should be removed")
        remove_ancestorid_from_defaults = True
        all_ok = False

    for backend in backends_to_check:
        log.info("Checking backend: %s", backend)

        # Check for ancestorid config entry (should not exist)
        if _index_config_exists(dse_ldif, backend, "ancestorid"):
            log.warning("  ancestorid - config entry exists (should be removed)")
            ancestorid_configs_to_remove.append(backend)
            all_ok = False

        # Check parentid and ancestorid indexes
        for index_name in ["parentid", "ancestorid"]:
            # Check for scan limits (should be removed)
            if _has_index_scan_limit(dse_ldif, backend, index_name):
                log.warning("  %s - has nsIndexIDListScanLimit (should be removed)", index_name)
                scan_limits_to_remove.append((backend, index_name))
                all_ok = False

            # Check disk ordering
            disk_ordering = _check_disk_ordering(db_dir, backend, index_name, dbscan_path, is_mdb, log)

            if disk_ordering == IndexOrdering.UNKNOWN:
                log.info("  %s - could not determine disk ordering, skipping", index_name)
                continue

            config_has_int_order = _has_integer_ordering_match(dse_ldif, backend, index_name)
            config_desc = "integer" if config_has_int_order else "lexicographic"
            log.info("  %s - config: %s, disk: %s",
                     index_name, config_desc, disk_ordering.value)

            # Both orderings are valid for parentid, but config must match disk.
            if index_name == "parentid":
                if config_has_int_order and disk_ordering == IndexOrdering.LEXICOGRAPHIC:
                    log.warning("  %s - MISMATCH: config has integerOrderingMatch but disk is lexicographic", index_name)
                    config_fixes.append((backend, index_name, "remove_mr"))
                    all_ok = False
                elif not config_has_int_order and disk_ordering == IndexOrdering.INTEGER:
                    log.warning("  %s - MISMATCH: config is lexicographic but disk has integer ordering", index_name)
                    config_fixes.append((backend, index_name, "add_mr"))
                    all_ok = False

    # Handle issues
    if not all_ok:
        if args.fix:
            log.info("Fixing issues...")

            # Remove ancestorid from cn=default indexes
            if remove_ancestorid_from_defaults:
                default_idx_dn = "cn=ancestorid,cn=default indexes,cn=config,cn=ldbm database,cn=plugins,cn=config"
                log.info("  Removing ancestorid from default indexes...")
                try:
                    dse_ldif.delete_dn(default_idx_dn)
                    log.info("  Removed ancestorid from default indexes")
                except Exception as e:
                    log.error("  Failed to remove ancestorid from default indexes: %s", e)
                    return False

            # Remove scan limits (only for indexes that won't be deleted)
            for backend, index_name in scan_limits_to_remove:
                # Skip ancestorid if we're going to delete the whole entry anyway
                if index_name == "ancestorid" and backend in ancestorid_configs_to_remove:
                    continue
                index_dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
                    index_name, backend
                )
                log.info("  Removing nsIndexIDListScanLimit from %s in backend %s...", index_name, backend)
                try:
                    dse_ldif.delete(index_dn, "nsIndexIDListScanLimit")
                    log.info("  Removed nsIndexIDListScanLimit from %s", index_name)
                except Exception as e:
                    log.error("  Failed to remove nsIndexIDListScanLimit from %s: %s", index_name, e)
                    return False

            # Remove ancestorid config entries from backends
            for backend in ancestorid_configs_to_remove:
                index_dn = "cn=ancestorid,cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(backend)
                log.info("  Removing ancestorid config entry from backend %s...", backend)
                try:
                    dse_ldif.delete_dn(index_dn)
                    log.info("  Removed ancestorid config entry from backend %s", backend)
                except Exception as e:
                    log.error("  Failed to remove ancestorid config from backend %s: %s", backend, e)
                    return False

            # Fix config-vs-disk ordering mismatches by adjusting config to match disk
            for backend, index_name, action in config_fixes:
                index_dn = "cn={},cn=index,cn={},cn=ldbm database,cn=plugins,cn=config".format(
                    index_name, backend
                )
                if action == "add_mr":
                    log.info("  Adding integerOrderingMatch to %s in backend %s...", index_name, backend)
                    try:
                        dse_ldif.add(index_dn, "nsMatchingRule", "integerOrderingMatch")
                        log.info("  Updated dse.ldif with integerOrderingMatch for %s", index_name)
                    except Exception as e:
                        log.error("  Failed to update dse.ldif for %s: %s", index_name, e)
                        return False
                elif action == "remove_mr":
                    log.info("  Removing integerOrderingMatch from %s in backend %s...", index_name, backend)
                    try:
                        dse_ldif.delete(index_dn, "nsMatchingRule", "integerOrderingMatch")
                        log.info("  Removed integerOrderingMatch from %s", index_name)
                    except Exception as e:
                        log.error("  Failed to remove integerOrderingMatch from %s: %s", index_name, e)
                        return False

            log.info("All issues fixed")
            return True
        else:
            log.info("Issues detected. Run with --fix to repair.")
            return False
    else:
        log.info("All checks passed - no issues found")
        return True


def create_parser(subcommands):
    db2index_parser = subcommands.add_parser('db2index', help="Initialise a reindex of the server database. The server must be stopped for this to proceed.", formatter_class=CustomHelpFormatter)
    db2index_parser.add_argument('backend', help="The backend to reindex. IE userRoot")
    db2index_parser.add_argument('--attr', action='append', help="An attribute to reindex. IE: --attr member --attr cn ...")
    db2index_parser.set_defaults(func=dbtasks_db2index)

    db2bak_parser = subcommands.add_parser('db2bak', help="Initialise a BDB backup of the database. The server must be stopped for this to proceed.", formatter_class=CustomHelpFormatter)
    db2bak_parser.add_argument('archive', help="The destination for the archive. This will be created during the db2bak process.",
                               nargs='?', default=None)
    db2bak_parser.add_argument('--watch', action='store_true', help='Watch the status of the db2bak task')
    db2bak_parser.set_defaults(func=dbtasks_db2bak)

    db2ldif_parser = subcommands.add_parser('db2ldif', help="Initialise an LDIF dump of the database. The server must be stopped for this to proceed.", formatter_class=CustomHelpFormatter)
    db2ldif_parser.add_argument('backend', help="The backend to output as an LDIF. IE userRoot")
    db2ldif_parser.add_argument('ldif', help="The path to the ldif output location.", nargs='?', default=None)
    db2ldif_parser.add_argument('--replication', help="Export replication information, suitable for importing on a new consumer or backups.",
                                default=False, action='store_true')
    # db2ldif_parser.add_argument('--include-changelog', help="Include the changelog as a separate LDIF file which will be named:  <ldif_file_name>_cl.ldif.  "
    #                                                         "This option also implies the '--replication' option is set.",
    #                             default=False, action='store_true')
    db2ldif_parser.add_argument('--encrypted', help="Export encrypted attributes", default=False, action='store_true')
    db2ldif_parser.add_argument('--watch', action='store_true', help='Watch the status of the db2ldif task')
    db2ldif_parser.set_defaults(func=dbtasks_db2ldif)

    dbverify_parser = subcommands.add_parser('dbverify', help="Perform a db verification. You should only do this at direction of support", formatter_class=CustomHelpFormatter)
    dbverify_parser.add_argument('backend', help="The backend to verify. IE userRoot")
    dbverify_parser.set_defaults(func=dbtasks_verify)

    bak2db_parser = subcommands.add_parser('bak2db', help="Restore a BDB backup of the database. The server must be stopped for this to proceed.", formatter_class=CustomHelpFormatter)
    bak2db_parser.add_argument('archive', help="The archive to restore. This will erase all current server databases.")
    bak2db_parser.add_argument('--watch', action='store_true', help='Watch the status of the bak2db task')
    bak2db_parser.set_defaults(func=dbtasks_bak2db)

    ldif2db_parser = subcommands.add_parser('ldif2db', help="Restore an LDIF dump of the database. The server must be stopped for this to proceed.", formatter_class=CustomHelpFormatter)
    ldif2db_parser.add_argument('backend', help="The backend to restore from an LDIF. IE userRoot")
    ldif2db_parser.add_argument('ldif', help="The path to the ldif to import")
    ldif2db_parser.add_argument('--encrypted', help="Import encrypted attributes", default=False, action='store_true')
    # ldif2db_parser.add_argument('--include-changelog', help="Include a replication changelog LDIF file if present.  It must be named like this in order for the import to include it:  <ldif_file_name>_cl.ldif.",
    #                            default=False, action='store_true')
    ldif2db_parser.add_argument('--watch', action='store_true', help='Watch the status of the ldif2db task')
    ldif2db_parser.set_defaults(func=dbtasks_ldif2db)

    backups_parser = subcommands.add_parser('backups', help="List backup's found in the server's default backup directory", formatter_class=CustomHelpFormatter)
    backups_parser.add_argument('--delete', nargs=1, help="Delete backup directory")
    backups_parser.set_defaults(func=dbtasks_backups)

    ldifs_parser = subcommands.add_parser('ldifs', help="List all the LDIF files located in the server's LDIF directory", formatter_class=CustomHelpFormatter)
    ldifs_parser.add_argument('--delete', nargs=1, help="Delete LDIF file")
    ldifs_parser.set_defaults(func=dbtasks_ldifs)

    index_check_parser = subcommands.add_parser('index-check',
        help="Check for index ordering mismatches (parentid/ancestorid). The server must be stopped.",
        formatter_class=CustomHelpFormatter)
    index_check_parser.add_argument('backend', nargs='?', default=None,
        help="Backend to check. If not specified, all backends are checked.")
    index_check_parser.add_argument('--fix', action='store_true', default=False,
        help="Fix mismatches by adjusting config to match on-disk data")
    index_check_parser.set_defaults(func=dbtasks_index_check)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---


from lib389._mapped_object import DSLdapObject


class RootDSE(DSLdapObject):
    """
    Check if the directory supports features or not.
    """
    def __init__(self, conn):
        """@param conn - a DirSrv instance """
        super(RootDSE, self).__init__(instance=conn)
        self._dn = ""

    def supported_sasl(self):
        return self.get_attr_vals_utf8('supportedSASLMechanisms')

    def available_sasl(self):
        return self.get_attr_vals_utf8('availableSASLMechanisms')

    def supports_sasl_gssapi(self):
        return self.present("supportedSASLMechanisms", 'GSSAPI')

    def supports_sasl_ldapssotoken(self):
        return self.present("supportedSASLMechanisms", "LDAPSSOTOKEN")

    def supports_sasl_plain(self):
        return self.present("supportedSASLMechanisms", "PLAIN")

    def supports_sasl_external(self):
        return self.present("supportedSASLMechanisms", "EXTERNAL")

    def supports_exop_whoami(self):
        return self.present("supportedExtension", "1.3.6.1.4.1.4203.1.11.3")

    def supports_exop_ldapssotoken_request(self):
        return self.present("supportedExtension", "2.16.840.1.113730.3.5.14")

    def supports_exop_ldapssotoken_revoke(self):
        return self.present("supportedExtension", "2.16.840.1.113730.3.5.16")

    def get_supported_ctrls(self):
        return self.get_attr_vals_utf8('supportedControl')




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import sys
import os
import ldap
from lib389.properties import (SER_LDAP_URL, SER_ROOT_DN, SER_LDAPI_ENABLED,
                               SER_LDAPI_SOCKET, SER_LDAPI_AUTOBIND)
from lib389._constants import DSRC_CONTAINER

MAJOR, MINOR, _, _, _ = sys.version_info

if MAJOR >= 3:
    import configparser


def dsrc_arg_concat(args, dsrc_inst):
    """
    Given a set of argparse args containing:

    instance
    binddn
    starttls

    and a dsrc_inst (as from dsrc_to_ldap)

    Using this, overlay the settings together to form an instance

    Generally, if dsrc_inst is None, we make a new instance.

    If dsrc_inst is populated we overlay args as needed on top.
    """
    if dsrc_inst is None:
        new_dsrc_inst = {
            'uri': args.instance,
            'basedn': args.basedn,
            'binddn': args.binddn,
            'bindpw': None,
            'saslmech': None,
            'tls_cacertdir': None,
            'tls_cert': None,
            'tls_key': None,
            'tls_reqcert': None,
            'starttls': args.starttls,
            'prompt': False,
            'pwdfile': None,
            'args': {}
        }
        # Now gather the args
        new_dsrc_inst['args'][SER_LDAP_URL] = new_dsrc_inst['uri']
        new_dsrc_inst['args'][SER_ROOT_DN] = new_dsrc_inst['binddn']
        if new_dsrc_inst['uri'][0:8] == 'ldapi://':
            new_dsrc_inst['args'][SER_LDAPI_ENABLED] = "on"
            new_dsrc_inst['args'][SER_LDAPI_SOCKET] = new_dsrc_inst['uri'][8:]
            new_dsrc_inst['args'][SER_LDAPI_AUTOBIND] = "on"

        # Make new
        return new_dsrc_inst
    # overlay things.
    if args.basedn is not None:
        dsrc_inst['basedn'] = args.basedn
    if args.binddn is not None:
        dsrc_inst['binddn'] = args.binddn
    if args.starttls is True:
        dsrc_inst['starttls'] = True
    return dsrc_inst


def _read_dsrc(path, log, case_sensetive=False):
    path = os.path.expanduser(path)
    log.debug("dsrc path: %s" % path)
    log.debug("dsrc container path: %s" % DSRC_CONTAINER)
    config = configparser.ConfigParser()
    if case_sensetive:
        config.optionxform = str
    # First read our container config if it exists
    # Then overlap the user config.
    config.read([DSRC_CONTAINER, path])

    log.debug("dsrc instances: %s" % config.sections())

    return config


def dsrc_to_ldap(path, instance_name, log):
    """
    Given a path to a file, return the required details for an instance.

    The file should be an ini file, and instance should identify a section.

    The ini file should have the content:

    [instance]
    uri = ldaps://hostname:port
    basedn = dc=example,dc=com
    people_rdn = ou=people
    groups_rdn = ou=groups
    binddn = uid=user,....
    saslmech = [EXTERNAL|PLAIN]
    tls_cacertdir = /path/to/cadir
    tls_cert = /path/to/user.crt
    tls_key = /path/to/user.key
    tls_reqcert = [never, hard, allow]
    starttls = [true, false]
    """
    config = _read_dsrc(path, log)
    server_id = instance_name

    # Do we have an instance name to work with?
    if instance_name is None:
        log.debug("No instance name provided")
        return None

    # Strip the prefix
    if instance_name.startswith("slapd-"):
        server_id = instance_name = instance_name.replace("slapd-", "", 1)

    if not config.has_section(instance_name):
        # instance_name does not have a prefix, but dsrc might, so add it
        instance_name = "slapd-" + instance_name
        if not config.has_section(instance_name):
            log.debug("dsrc no such section: %s" % instance_name)
            return None

    dsrc_inst = {}
    dsrc_inst['args'] = {}

    # Read all the values
    dsrc_inst['uri'] = config.get(instance_name, 'uri', fallback=f"ldapi://%2fvar%2frun%2fslapd-{server_id}.socket")
    dsrc_inst['basedn'] = config.get(instance_name, 'basedn', fallback=None)
    dsrc_inst['people_rdn'] = config.get(instance_name, 'people_rdn', fallback=None)
    dsrc_inst['groups_rdn'] = config.get(instance_name, 'groups_rdn', fallback=None)
    dsrc_inst['binddn'] = config.get(instance_name, 'binddn', fallback=None)
    dsrc_inst['saslmech'] = config.get(instance_name, 'saslmech', fallback=None)
    if dsrc_inst['saslmech'] is not None and dsrc_inst['saslmech'] not in ['EXTERNAL', 'PLAIN']:
        raise ValueError("%s [%s] saslmech must be one of EXTERNAL or PLAIN" % (path, instance_name))

    dsrc_inst['tls_cacertdir'] = config.get(instance_name, 'tls_cacertdir', fallback=None)
    # At this point, we should check if the provided cacertdir is indeed, a dir. This can be a cause
    # of basic failures and confusion.
    if dsrc_inst['tls_cacertdir'] is not None and (os.path.exists(dsrc_inst['tls_cacertdir']) == False or os.path.isdir(dsrc_inst['tls_cacertdir']) == False):
        log.warning("Warning: dsrc tls_cacertdir path may not exist, or is not a directory")
        log.warning("Warning: This should be a directory, and you must run '/usr/bin/c_rehash path' for it to be a valid CA store")

    dsrc_inst['tls_cert'] = config.get(instance_name, 'tls_cert', fallback=None)
    dsrc_inst['tls_key'] = config.get(instance_name, 'tls_key', fallback=None)
    dsrc_inst['tls_reqcert'] = config.get(instance_name, 'tls_reqcert', fallback=None)
    if dsrc_inst['tls_reqcert'] == 'never':
        dsrc_inst['tls_reqcert'] = ldap.OPT_X_TLS_NEVER
    elif dsrc_inst['tls_reqcert'] == 'allow':
        dsrc_inst['tls_reqcert'] = ldap.OPT_X_TLS_ALLOW
    elif dsrc_inst['tls_reqcert'] == 'hard':
        dsrc_inst['tls_reqcert'] = ldap.OPT_X_TLS_HARD
    elif dsrc_inst['tls_reqcert'] is None:
        # Use system value
        pass
    else:
        raise ValueError("dsrc tls_reqcert value invalid. %s [%s] tls_reqcert should be one of never, allow or hard" % (instance_name, path))
    dsrc_inst['starttls'] = config.getboolean(instance_name, 'starttls', fallback=False)
    dsrc_inst['pwdfile'] = None
    dsrc_inst['prompt'] = False
    # Now gather the args
    dsrc_inst['args']['server-id'] = server_id
    dsrc_inst['args'][SER_LDAP_URL] = dsrc_inst['uri']
    dsrc_inst['args'][SER_ROOT_DN] = dsrc_inst['binddn']
    if dsrc_inst['uri'][0:8] == 'ldapi://':
        dsrc_inst['args'][SER_LDAPI_ENABLED] = "on"
        dsrc_inst['args'][SER_LDAPI_SOCKET] = dsrc_inst['uri'][8:]
        dsrc_inst['args'][SER_LDAPI_AUTOBIND] = "on"

    # Return the dict.
    log.debug("dsrc completed with %s" % dsrc_inst)
    return dsrc_inst


def dsrc_to_repl_monitor(path, log):
    """
    Given a path to a file, return the required details for an instance.

    The connection values for monitoring other not connected topologies. The format:
    'host:port:binddn:bindpwd'. You can use regex for host and port. You can set bindpwd
    to * and it will be requested at the runtime.

    If a host:port is assigned an alias, then the alias instead of host:port will be
    displayed in the output. The format: "alias=host:port".

    The file should be an ini file, and instance should identify a section.

    The ini file should have the content:

    [repl-monitor-connections]
    connection1 = server1.example.com:38901:cn=Directory manager:*
    connection2 = server2.example.com:38902:cn=Directory manager:[~/pwd.txt]
    connection3 = hub1.example.com:.*:cn=Directory manager:password

    [repl-monitor-aliases]
    M1 = server1.example.com:38901
    M2 = server2.example.com:38902
    """

    config = _read_dsrc(path, log, case_sensetive=True)
    dsrc_repl_monitor = {"connections": None,
                         "aliases": None}

    if config.has_section("repl-monitor-connections"):
        dsrc_repl_monitor["connections"] = [conn for _, conn in config.items("repl-monitor-connections")]

    if config.has_section("repl-monitor-aliases"):
        dsrc_repl_monitor["aliases"] = {alias: inst for alias, inst in config.items("repl-monitor-aliases")}

    log.debug(f"dsrc completed with {dsrc_repl_monitor}")
    return dsrc_repl_monitor



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Brooker classes to organize ldap methods.
   Stuff is split in classes, like:
   * Replica
   * Backend
   * Suffix

   You will access this from:
   DirSrv.backend.methodName()
"""

import copy
import ldap
from lib389._constants import *
from lib389 import Entry
from lib389._mapped_object import DSLdapObject, DSLdapObjects
from lib389.utils import ensure_bytes, selinux_label_port,  selinux_present
from lib389.lint import (
    DSCLE0002, DSCLE0003, DSCLE0004, DSCLE0005, DSCLE0006, DSELE0001
)
from lib389._mapped_object_lint import (
    lint_get_attr_val,
    lint_get_attr_val_utf8,
    lint_get_attr_val_utf8_l,
)

class Config(DSLdapObject):
    """
        Manage "cn=config" tree, including:
        - enable SSL
        - set access and error logging
        - get and set "cn=config" attributes
    """
    def __init__(self, conn, dn=None):
        """@param conn - a DirSrv instance """
        super(Config, self).__init__(instance=conn)
        self._dn = DN_CONFIG
        # self._instance = conn
        self.log = conn.log
        config_compare_exclude = [
            'nsslapd-ldapifilepath',
            'nsslapd-accesslog',
            'nsslapd-auditfaillog',
            'nsslapd-ldifdir',
            'nsslapd-errorlog',
            'nsslapd-securitylog',
            'nsslapd-instancedir',
            'nsslapd-lockdir',
            'nsslapd-bakdir',
            'nsslapd-schemadir',
            'nsslapd-auditlog',
            'nsslapd-rootpw',
            'nsslapd-workingdir',
            'nsslapd-certdir'
        ]
        self._compare_exclude  = self._compare_exclude + config_compare_exclude
        self._rdn_attribute = 'cn'

    @property
    def dn(self):
        return DN_CONFIG

    @property
    def rdn(self):
        return DN_CONFIG

    def replace(self, key, value):
        if selinux_present() and (key.lower() == 'nsslapd-secureport' or key.lower() == 'nsslapd-port'):
            # Get old port and remove label
            old_port = self.get_attr_val_utf8(key)
            self.log.debug("Removing old port's SELinux label...")
            selinux_label_port(old_port, remove_label=True)
            self.log.debug("Setting new port's SELinux label...")
            selinux_label_port(value)
        super(Config, self).replace(key,  value)

    def _alter_log_enabled(self, service, state):
        if service not in ('access', 'error', 'audit'):
            self._log.error('Attempted to enable invalid log service "%s"' % service)
        service = 'nsslapd-%slog-logging-enabled' % service
        self._log.debug('Setting log %s to %s' % (service, state))
        self.set(service, state)

    def enable_log(self, service):
        """Enable a logging service in the 389ds instance.
        @param service - The logging service to enable. Can be one of 'access',
                         'error' or 'audit'.

        ex. enable_log('audit')
        """
        self._alter_log_enabled(service, 'on')

    def disable_log(self, service):
        """Disable a logging service in the 389ds instance.
        @param service - The logging service to Disable. Can be one of 'access'
                         , 'error' or 'audit'.

        ex. disable_log('audit')
        """
        self._alter_log_enabled(service, 'off')

    def loglevel(self, vals=(ErrorLog.DEFAULT,), service='error', update=False):
        """Set the access or error log level.

        :param vals: a list of log level codes (eg. lib389.ErrorLogLevelLOG_*)
                      defaults to LOG_DEFAULT
        :type vals: list
        :param service: 'access' or 'error'. There is no 'audit' log level.
                         use enable_log or disable_log.
        :type service: str
        :param update: False for replace (default), True for update
        :type update: bool

        ex. loglevel([lib389.ErrorLogLevel.DEFAULT, lib389.ErrorLogLevel.ENTRY_PARSER])
        """
        if service not in ('access', 'error'):
            self._log.error('Attempted to set level on invalid log service "%s"' % service)
        service = 'nsslapd-%slog-level' % service
        assert len(vals) > 0, "set at least one log level"
        tot = 0
        for v in vals:
            tot |= v

        if update:
            old = int(self.get(service))
            tot |= old
            self._log.debug("Update %s value: %r -> %r" % (service, old, tot))
        else:
            self._log.debug("Replace %s with value: %r" % (service, tot))

        self.set(service, str(tot))
        return tot

    def logbuffering(self, state=True):
        if state:
            value = 'on'
        else:
            value = 'off'

        self.set('nsslapd-accesslog-logbuffering', value)

    def reset(self, key):
        self.set(key, None, action=ldap.MOD_DELETE)

    # THIS WILL BE SPLIT OUT TO ITS OWN MODULE
    def enable_ssl(self, secport=636, secargs=None):
        """Configure SSL support into cn=encryption,cn=config.

            secargs is a dict like {
                'nsSSLPersonalitySSL': 'Server-Cert'
            }
        """
        self._log.debug("config.enable_ssl is deprecated! Use RSA, Encryption instead!")
        self._log.debug("configuring SSL with secargs:%r" % secargs)
        secargs = secargs or {}

        dn_enc = 'cn=encryption,cn=config'
        ciphers = ('-rsa_null_md5,+rsa_rc4_128_md5,+rsa_rc4_40_md5,'
                   '+rsa_rc2_40_md5,+rsa_des_sha,+rsa_fips_des_sha,'
                   '+rsa_3des_sha,+rsa_fips_3des_sha,+tls_rsa_export1024'
                   '_with_rc4_56_sha,+tls_rsa_export1024_with_des_cbc_sha')
        mod = [(ldap.MOD_REPLACE, 'nsSSL3', secargs.get('nsSSL3', 'on')),
               (ldap.MOD_REPLACE, 'nsSSLClientAuth',
                secargs.get('nsSSLClientAuth', 'allowed')),
               (ldap.MOD_REPLACE, 'nsSSL3Ciphers', secargs.get('nsSSL3Ciphers',
                ciphers))]
        self._instance.modify_s(dn_enc, mod)

        dn_rsa = 'cn=RSA,cn=encryption,cn=config'
        e_rsa = Entry(dn_rsa)
        e_rsa.update({
            'objectclass': ['top', 'nsEncryptionModule'],
            'nsSSLPersonalitySSL': secargs.get('nsSSLPersonalitySSL',
                                               'Server-Cert'),
            'nsSSLToken': secargs.get('nsSSLToken', 'internal (software)'),
            'nsSSLActivation': secargs.get('nsSSLActivation', 'on')
        })
        try:
            self._instance.add_s(e_rsa)
        except ldap.ALREADY_EXISTS:
            pass

        mod = [
            (ldap.MOD_REPLACE,
             'nsslapd-security',
             secargs.get('nsslapd-security', 'on')),
            (ldap.MOD_REPLACE,
             'nsslapd-ssl-check-hostname',
             secargs.get('nsslapd-ssl-check-hostname', 'off')),
            (ldap.MOD_REPLACE,
             'nsslapd-secureport',
             str(secport))
        ]
        self._log.debug("trying to modify %r with %r" % (DN_CONFIG, mod))
        self._instance.modify_s(DN_CONFIG, mod)

        fields = 'nsslapd-security nsslapd-ssl-check-hostname'.split()
        return self._instance.getEntry(DN_CONFIG, attrlist=fields)

    @classmethod
    def lint_uid(cls):
        return 'config'

    def _lint_passwordscheme(self):
        allowed_schemes = ['PBKDF2-SHA512', 'PBKDF2_SHA256', 'PBKDF2_SHA512', 'GOST_YESCRYPT']
        u_password_scheme = lint_get_attr_val_utf8(self, 'passwordStorageScheme')
        u_root_scheme = lint_get_attr_val_utf8(self, 'nsslapd-rootpwstoragescheme')
        if u_root_scheme not in allowed_schemes:
            report = copy.deepcopy(DSCLE0002)
            report['detail'] = report['detail'].replace('SCHEME', u_root_scheme)
            report['detail'] = report['detail'].replace('CONFIG', 'nsslapd-rootpwstoragescheme')
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['fix'] = report['fix'].replace('CONFIG', 'nsslapd-rootpwstoragescheme')
            report['check'] = "config:passwordscheme"
            yield report
        if u_password_scheme not in allowed_schemes:
            report = copy.deepcopy(DSCLE0002)
            report['detail'] = report['detail'].replace('SCHEME', u_password_scheme)
            report['detail'] = report['detail'].replace('CONFIG', 'passwordStorageScheme')
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['fix'] = report['fix'].replace('CONFIG', 'passwordStorageScheme')
            report['check'] = "config:passwordscheme"
            yield report

    def _lint_unauth_binds(self):
        # Allow unauthenticated binds
        unauthbinds = lint_get_attr_val_utf8_l(self, 'nsslapd-allow-unauthenticated-binds')
        if unauthbinds == "on":
            report = copy.deepcopy(DSCLE0003)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['check'] = "config:unauthorizedbinds"
            yield report

    def _lint_accesslog_buffering(self):
        # access log buffering
        buffering = lint_get_attr_val_utf8_l(self, 'nsslapd-accesslog-logbuffering')
        if buffering == "off":
            report = copy.deepcopy(DSCLE0004)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['check'] = "config:accesslogbuffering"
            yield report

    def _lint_auditlog_buffering(self):
        # audit log buffering
        enabled = lint_get_attr_val_utf8_l(self, 'nsslapd-auditlog-logging-enabled')
        buffering = lint_get_attr_val_utf8_l(self, 'nsslapd-auditlog-logbuffering')
        if enabled == "on" and buffering == "off":
            report = copy.deepcopy(DSCLE0006)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['check'] = "config:auditlogbuffering"
            yield report

    def _lint_securitylog_buffering(self):
        # security log buffering
        enabled = lint_get_attr_val_utf8_l(self, 'nsslapd-securitylog-logging-enabled')
        buffering = lint_get_attr_val_utf8_l(self, 'nsslapd-securitylog-logbuffering')
        if enabled == "on" and buffering == "off":
            report = copy.deepcopy(DSCLE0005)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['check'] = "config:securitylogbuffering"
            yield report

    def disable_plaintext_port(self):
        """
        Configure the server to not-provide the plaintext port.
        """
        self.set('nsslapd-port', '0')

    def enable_plaintext_port(self, port):
        """
        Configure the server to provide the plaintext port on the specified port number.
        """
        self.set('nsslapd-port', port)


class Encryption(DSLdapObject):
    """
        Manage "cn=encryption,cn=config" tree, including:
        - ssl ciphers
        - ssl / tls levels
    """
    def __init__(self, conn, dn=None):
        """@param conn - a DirSrv instance """
        assert dn is None  # compatibility with Config class
        super(Encryption, self).__init__(instance=conn)
        self._dn = 'cn=encryption,%s' % DN_CONFIG
        self._create_objectclasses = ['top', 'nsEncryptionConfig']
        # Once created, don't allow it's removal
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._protected = True

    def create(self, rdn=None, properties={'cn': 'encryption', 'nsSSLClientAuth': 'allowed'}):
        if rdn is not None:
            self._log.debug("dn on cn=encryption is not None. This is a mistake.")
        super(Encryption, self).create(properties=properties)

    @classmethod
    def lint_uid(cls):
        return 'encryption'

    def _lint_check_tls_version(self):
        tls_min = lint_get_attr_val(self, 'sslVersionMin')
        if tls_min is not None and tls_min < ensure_bytes('TLS1.1'):
            report = copy.deepcopy(DSELE0001)
            report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
            report['check'] = "encryption:check_tls_version"
            yield report

    @property
    def ciphers(self):
        """List of requested ciphers.

        Each is represented by a string, either of:
        - "+all" or "-all"
        - TLS cipher RFC name, prefixed with either "+" or "-"

        Optionally, first element may be a string "default".

        :returns: list of str
        """
        val = self.get_attr_val_utf8('nsSSL3Ciphers')
        if val:
            return val.split(',')
        else:
            return ['default']

    @ciphers.setter
    def ciphers(self, ciphers):
        """List of requested ciphers.

        :param ciphers: Ciphers to enable
        :type ciphers: list of str
        """
        self.set('nsSSL3Ciphers', ','.join(ciphers))
        self._log.info('Remember to restart the server to apply the new cipher set.')
        self._log.info('Some ciphers may be disabled anyway due to allowWeakCipher attribute.')

    def _get_listed_ciphers(self, attr):
        """Remove features of ciphers that come after first :: occurence."""
        return [c[:c.index('::')] for c in self.get_attr_vals_utf8(attr)]

    @property
    def enabled_ciphers(self):
        """List currently enabled ciphers.

        :returns: list of str
        """
        return self._get_listed_ciphers('nsSSLEnabledCiphers')

    @property
    def supported_ciphers(self):
        """List currently supported ciphers.

        :returns: list of str
        """
        return self._get_listed_ciphers('nsSSLSupportedCiphers')

    def _check_ciphers_supported(self, ciphers):
        good = True
        for c in ciphers:
            if c not in self.supported_ciphers:
                self._log.warn(f'Cipher {c} is not supported.')
                good = False
        return good

    def change_ciphers(self, mode, ciphers):
        """Enable or disable ciphers of the nsSSL3Ciphers attribute.

        :param mode: '+'/'-' string to enable/disable the ciphers
        :type mode: str
        :param ciphers: List of ciphers to enable/disable
        :type ciphers: list of string

        :returns: False if some cipher is not supported
        """
        if ('default' in ciphers) or 'all' in ciphers:
            raise NotImplementedError('Processing "default" and "all" is not implemented.')
        if not self._check_ciphers_supported(ciphers):
            return False

        if mode == '+':
            to_change = [c for c in ciphers if c not in self.enabled_ciphers]
        elif mode == '-':
            to_change = [c for c in ciphers if c in self.enabled_ciphers]
        else:
            raise ValueError('Incorrect mode. Use - or + sign.')
        if len(to_change) != len(ciphers):
            self._log.info(
                ('Applying changes only for the following ciphers, the rest is up to date. '
                 'If this does not seem to be correct, please make sure the effective '
                 'set of enabled ciphers is up to date with configured ciphers '
                 '- a server restart is needed for these to be applied.\n'
                 f'... {to_change}'))
        cleaned = [c for c in self.ciphers if c[1:] not in to_change]
        self.ciphers = cleaned + list(map(lambda c: mode + c, to_change))


class RSA(DSLdapObject):
    """
        Manage the "cn=RSA,cn=encryption,cn=config" object
        - Set the certificate name
        - Database path
        - ssl token name
    """
    def __init__(self, conn, dn=None):
        """@param conn - a DirSrv instance """
        assert dn is None  # compatibility with Config class
        super(RSA, self).__init__(instance=conn)
        self._dn = 'cn=RSA,cn=encryption,%s' % DN_CONFIG
        self._create_objectclasses = ['top', 'nsEncryptionModule']
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        # Once we create it, don't remove it
        self._protected = True

    def _validate(self, rdn, properties, basedn):
        (dn, valid_props) = super(RSA, self)._validate(rdn, properties, basedn)
        # Ensure that dn matches self._dn
        assert(self._dn == dn)
        return (dn, valid_props)

    def create(self, rdn=None, properties={'cn': 'RSA', 'nsSSLPersonalitySSL': 'Server-Cert', 'nsSSLActivation': 'on', 'nsSSLToken': 'internal (software)'}):
        # Is this the best way to for the dn?
        if rdn is not None:
            self._log.debug("dn on cn=Rsa create request is not None. This is a mistake.")
        # Our self._dn is already set, no need for rdn.
        super(RSA, self).create(properties=properties)


class EncryptionModule(DSLdapObject):
    """EncryptionModule DSLdapObject with:
    - must attributes = ['cn']
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(EncryptionModule, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn', 'nsSSLPersonalitySSL', 'nsSSLActivation', 'nsSSLToken']
        self._create_objectclasses = ['top', 'nsEncryptionModule']
        self._protected = False


class EncryptionModules(DSLdapObjects):
    """DSLdapObjects that represents Encryption modules

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance):
        super(EncryptionModules, self).__init__(instance=instance)
        self._objectclasses = ['nsEncryptionModule']
        self._filterattrs = ['cn', 'nsSSLPersonalitySSL']
        self._childobject = EncryptionModule
        self._basedn = 'cn=encryption,cn=config'

    def create(self, properties=None):
        if properties is None:
            properties = {}
        token_set = False
        for k, v in properties.items():
            if k.lower() == 'nsssltoken':
                token_set = True
                break

        if not token_set:
            # Set the token name by default since it doesn't change
            properties['nsSSLToken'] = 'internal (software)'

        # Our self._dn is already set, no need for rdn.
        super(EncryptionModules, self).create(rdn=None, properties=properties)


class CertmapLegacy(object):
    """
    Manage certificate mappings in Directory Server

    This is based on the old certmap fileformat. As a result, the interface is rather
    crude, until we re-write to use dynamic mappings.
    """
    def __init__(self, conn):
        self._instance = conn
        pass

    def reset(self):
        """
        Reset the certmap to empty.
        """
        certmap = os.path.join(self._instance.get_config_dir(), 'certmap.conf')
        with open(certmap, 'w') as f:
            f.write('# LDAP Certificate mappings \n')

    def _parse_maps(self, maps):
        certmaps = {}

        for l in maps:
            if l.startswith('certmap'):
                # Line matches format of: certmap name issuer
                (_, name, issuer) = l.split(None, 2)
                certmaps[name] = {
                    'DNComps': None,
                    'FilterComps': None,
                    'VerifyCert': None,
                    'CmapLdapAttr': None,
                    'Library': None,
                    'InitFn': None,
                }
                certmaps[name]['issuer'] = issuer
            else:
                # The line likely is:
                # name:property [value]
                (name, pav) = l.split(':')
                pavs = pav.split(None, 1)
                if len(pavs) == 1:
                    # We have an empty property
                    certmaps[name][pav] = ''
                else:
                    # We clearly have a value.
                    if pavs[0] == 'DNComps' or pavs[0] == 'FilterComps':
                        # These two are comma sep lists
                        values = [w.split for w in pavs[1].split(',')]
                        certmaps[name][pavs[0]] = values
                    else:
                        certmaps[name][pavs[0]] = pavs[1]
        return certmaps

    def list(self):
        """
        Parse and list current certmaps.
        """
        certmap = os.path.join(self._instance.get_config_dir(), 'certmap.conf')
        maps = []
        with open(certmap, 'r') as f:
            for line in f.readlines():
                s_line = line.strip()
                if not s_line.startswith('#'):
                    content = s_line.split('#')[0]
                    if content != '':
                        maps.append(content)
        certmaps = self._parse_maps(maps)
        return certmaps

    def set(self, certmaps):
        """
        Take a dict of certmaps and write them out.
        """
        output = ""
        for name in certmaps:
            certmap = certmaps[name]
            output += "certmap %s %s\n" % (name, certmap['issuer'])
            for v in ['DNComps', 'FilterComps']:
                if certmap[v] == None:
                    output += "# %s:%s\n" % (name, v)
                elif certmap[v] == '':
                    output += "%s:%s\n" % (name, v)
                else:
                    output += "%s:%s %s\n" % (name, v, ', '.join(certmap[v]))
            for v in ['VerifyCert', 'CmapLdapAttr', 'Library', 'InitFn']:
                if certmap[v] == None:
                    output += "# %s:%s\n" % (name, v)
                else:
                    output += "%s:%s %s\n" % (name, v, certmap[v])
        # Now write it out
        certmap = os.path.join(self._instance.get_config_dir(), 'certmap.conf')
        if not os.access(certmap, os.W_OK):
            os.chmod(certmap, 0o660)
        with open(certmap, 'w') as f:
            f.write(output)
        if os.access(certmap, os.W_OK):
            os.chmod(certmap, 0o440)


class LDBMConfig(DSLdapObject):
    """
        Manage "cn=config,cn=ldbm database,cn=plugins,cn=config" including:
        - Performance related tunings
        - DB backend settings

        :param instance: An instance
        :type instance: lib389.DirSrv
    """

    def __init__(self, conn):
        super(LDBMConfig, self).__init__(instance=conn)
        self._dn = DN_CONFIG_LDBM
        # config_compare_exclude = []
        self._rdn_attribute = 'cn'
        self._protected = True


class BDB_LDBMConfig(DSLdapObject):
    """
        Manage "cn=bdb,cn=config,cn=ldbm database,cn=plugins,cn=config" including:
        - Performance related tunings
        - BDB specific DB backend settings

        :param instance: An instance
        :type instance: lib389.DirSrv
    """

    def __init__(self, conn):
        super(BDB_LDBMConfig, self).__init__(instance=conn)
        self._dn = DN_CONFIG_LDBM_BDB
        self._config_compare_exclude = []
        self._rdn_attribute = 'cn'
        self._protected = True

class LMDB_LDBMConfig(DSLdapObject):
    """
        Manage "cn=mdb,cn=config,cn=ldbm database,cn=plugins,cn=config" including:
        - Performance related tunings
        - LMDB specific DB backend settings

        :param instance: An instance
        :type instance: lib389.DirSrv
    """

    def __init__(self, conn):
        super(LMDB_LDBMConfig, self).__init__(instance=conn)
        self._dn = DN_CONFIG_LDBM_LMDB
        self._config_compare_exclude = []
        self._rdn_attribute = 'cn'
        self._protected = True




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016-2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import pytest

from lib389.tests.topologies import topology_st
from lib389.plugins import USNPlugin
from lib389.rootdse import RootDSE
from lib389._mapped_object import DSLdapObjects
from lib389.backend import Backend
from lib389.tests.plugins.utils import create_test_user, create_test_ou, delete_objects
from lib389._constants import DEFAULT_SUFFIX, DN_LDBM
from lib389.properties import BACKEND_NAME, BACKEND_SUFFIX


def reset_USN(plugin, objects):
    """
    Delete objects containing an entryusn attr, clean USN tombstone entries
    and restart the instance. This makes USN to reset lastusn back to -1.
    """
    delete_objects(objects)
    task = plugin.cleanup(suffix=DEFAULT_SUFFIX)
    task.wait()
    plugin._instance.restart()


@pytest.fixture(scope="module")
def plugin(request):
    topology = topology_st(request)
    plugin = USNPlugin(topology.standalone)
    return plugin


def test_usn_enable_disable(plugin):
    """
    Test that the plugin doesn't do anything while disabled, but stores
    entryusn values properly when enabled.

    NOTICE: This test case leaves the plugin enabled for the following tests.
    """
    # assert plugin is disabled (by default)
    assert plugin.status() == False

    root_dse = RootDSE(plugin._instance)
    # create a function for returning lastusn every time is called
    lastusn = lambda : root_dse.get_attr_val_int("lastusn;userroot")

    user1 = create_test_user(plugin._instance)

    # assert no entryusn,lastusn values were created while the plugin was disabled
    assert not "entryusn" in user1.get_all_attrs()
    assert not "lastusn;userroot" in root_dse.get_all_attrs()

    # enable the plugin and restart the server for the action to take effect
    plugin.enable()
    plugin._instance.restart()
    assert plugin.status() == True

    assert lastusn() == -1
    user2 = create_test_user(plugin._instance)
    # assert that a new entry now contains the entryusn value
    assert user2.get_attr_val_int("entryusn") == lastusn() == 0

    # assert that USNs are properly assigned after any write operation
    # write operations include add, modify, modrdn and delete operations
    user3 = create_test_user(plugin._instance)
    assert user3.get_attr_val_int("entryusn") == lastusn() == 1
    user2.delete()
    assert lastusn() == 2
    user3.replace('sn', 'another surname')
    assert user3.get_attr_val_int("entryusn") == lastusn() == 3

    # reset USN for subsequent test cases
    reset_USN(plugin, [user1, user3])

def test_usn_local_mode(plugin):
    """
    Test that when USN is operating in local mode, each backend has an instance
    of the USN Plug-in with a USN counter specific to that backend database.
    """
    # assert local mode
    assert not plugin.is_global_mode_set()

    ou_value = "People2"
    ou_suffix = DEFAULT_SUFFIX

    # create a new backend
    plugin._instance.backends.create(
        None, properties={
            BACKEND_NAME: "People2Data",
            BACKEND_SUFFIX: "ou=" + ou_value + "," + ou_suffix,
    })

    # create a new sub-suffix stored in the new backend
    ou2 = create_test_ou(plugin._instance, ou=ou_value, suffix=ou_suffix)

    root_dse = RootDSE(plugin._instance)
    lastusn_b1 = lambda : root_dse.get_attr_val_int("lastusn;userroot")
    lastusn_b2 = lambda : root_dse.get_attr_val_int("lastusn;people2data")

    assert lastusn_b1() == lastusn_b2() == -1

    # add a user in the default backend; trigger the plugin
    user_b1 = create_test_user(plugin._instance)
    user_b1.replace('sn', 'surname2')
    user_b1.replace('sn', 'surname3')

    # add a user in the new backend
    user_b2 = create_test_user(plugin._instance, suffix=ou2.dn)

    # assert the USN counter is different for each backend
    assert user_b1.get_attr_val_int("entryusn") == lastusn_b1() == 2
    assert user_b2.get_attr_val_int("entryusn") == lastusn_b2() == 0

    # reset USN for subsequent test cases
    b = Backend(plugin._instance, dn="cn=people2data," + DN_LDBM)
    reset_USN(plugin, [user_b1, user_b2, ou2, b])

def test_tombstone_cleanup(plugin):
    """
    Assert that the USN plugin removes tombstone entries when the cleanup task
    is run. Test removal for a specific backend, a specific suffix, and up to
    a given USN number.
    """
    # create a new backend and a new sub-suffix stored in the new backend
    ou_value = "People3"
    ou_suffix = DEFAULT_SUFFIX
    plugin._instance.backends.create(
        None, properties={
            BACKEND_NAME: "People3Data",
            BACKEND_SUFFIX: "ou=" + ou_value + "," + ou_suffix,
    })

    ou2 = create_test_ou(plugin._instance, ou=ou_value, suffix=ou_suffix)

    tombstones_b1 = DSLdapObjects(plugin._instance)
    tombstones_b1._basedn = "ou=People," + DEFAULT_SUFFIX
    tombstones_b1._objectclasses = ['nsTombstone']

    tombstones_b2 = DSLdapObjects(plugin._instance)
    tombstones_b2._basedn = ou2.dn
    tombstones_b2._objectclasses = ['nsTombstone']

    root_dse = RootDSE(plugin._instance)
    lastusn_b1 = lambda : root_dse.get_attr_val_int("lastusn;userroot")
    assert lastusn_b1() == -1

    user1_b1 = create_test_user(plugin._instance)
    user2_b1 = create_test_user(plugin._instance)
    user3_b1 = create_test_user(plugin._instance)

    user1_b2 = create_test_user(plugin._instance, suffix=ou2.dn)
    user2_b2 = create_test_user(plugin._instance, suffix=ou2.dn)

    # assert no tombstones exist at this point
    assert not tombstones_b1.list()
    assert not tombstones_b2.list()

    # create 3 tombstone entries on default backend
    user1_b1.delete()
    user2_b1.delete()
    user3_b1.delete()

    # assert there are 3 tombstone entries indeed on default backend
    assert len(tombstones_b1.list()) == 3
    assert not tombstones_b2.list()

    assert lastusn_b1() == 5

    # remove all tombstone entries from default backend, with a USN value up to 4
    task = plugin.cleanup(suffix=DEFAULT_SUFFIX, max_usn=lastusn_b1()-1)
    task.wait()

    # assert all tombstone entries were deleted but the last one on default backend
    assert len(tombstones_b1.list()) == 1
    assert not tombstones_b2.list()

    # create 2 tombstone entries on new backend
    user1_b2.delete()
    user2_b2.delete()

    # assert there are 2 tombstone entries indeed on new backend
    assert len(tombstones_b2.list()) == 2
    assert len(tombstones_b1.list()) == 1

    # remove all tombstone entries from ou2 suffix
    task = plugin.cleanup(suffix=ou2.dn)
    task.wait()

    # assert there are no tombstone entries stored on ou2 suffix
    assert not tombstones_b2.list()
    assert len(tombstones_b1.list()) == 1

    # reset USN for subsequent test cases
    b = Backend(plugin._instance, dn="cn=people3data," + DN_LDBM)
    reset_USN(plugin, [ou2, b])

def test_usn_global_mode(plugin):
    """
    Test that when USN is operating in global mode, there is a global instance
    of the USN Plug-in with a global USN counter that applies to changes made
    to the entire directory (all backends).
    """
    plugin.enable_global_mode()
    plugin._instance.restart()
    assert plugin.is_global_mode_set()

    # create a new backend and a new sub-suffix stored in the new backend
    ou_value = "People4"
    ou_suffix = DEFAULT_SUFFIX
    plugin._instance.backends.create(
        None, properties={
            BACKEND_NAME: "People4Data",
            BACKEND_SUFFIX: "ou=" + ou_value + "," + ou_suffix,
    })

    root_dse = RootDSE(plugin._instance)
    assert "lastusn" in root_dse.get_all_attrs()

    ou2 = create_test_ou(plugin._instance, ou=ou_value, suffix=ou_suffix)
    assert ou2.get_attr_val_int("entryusn") == 0

    # add a user in the default backend
    user_b1 = create_test_user(plugin._instance)
    # add a user in the new backend
    user_b2 = create_test_user(plugin._instance, suffix=ou2.dn)

    # assert that a global USN counter is used for all backends
    assert user_b1.get_attr_val_int("entryusn") == 1
    assert user_b2.get_attr_val_int("entryusn") == 2

    # reset USN for subsequent test cases
    b = Backend(plugin._instance, dn="cn=people4data," + DN_LDBM)
    reset_USN(plugin, [user_b1, user_b2, ou2, b])



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest
import ldap

from lib389._constants import DEFAULT_SUFFIX, INSTALL_LATEST_CONFIG

from lib389.cli_conf.backend import backend_create
from lib389.cli_idm.initialise import initialise
from lib389.cli_idm.group import get, create, delete, members, add_member, remove_member
from lib389.cli_idm.user import create as create_user

from lib389.cli_base import LogCapture, FakeArgs
from lib389.tests.cli import topology_be_latest as topology

from lib389.utils import ds_is_older
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")

# Topology is pulled from __init__.py
def test_group_tasks(topology):
    # First check that our test group isn't there:
    topology.logcap.flush()
    g_args = FakeArgs()
    g_args.selector = 'testgroup'
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        get(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)

    # Create a group
    topology.logcap.flush()
    g_args.cn = 'testgroup'
    create(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("Successfully created testgroup"))

    # Assert it exists
    topology.logcap.flush()
    g_args = FakeArgs()
    g_args.selector = 'testgroup'
    get(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("dn: cn=testgroup,ou=groups,dc=example,dc=com"))

    # Add a user
    topology.logcap.flush()
    u_args = FakeArgs()
    u_args.uid = 'testuser'
    u_args.cn = 'Test User'
    u_args.displayName = 'Test User'
    u_args.homeDirectory = '/home/testuser'
    u_args.uidNumber = '5000'
    u_args.gidNumber = '5000'
    create_user(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, u_args)
    assert(topology.logcap.contains("Successfully created testuser"))

    # Add them to the group as a member
    topology.logcap.flush()
    g_args.cn = "testgroup"
    g_args.dn = "uid=testuser,ou=people,dc=example,dc=com"
    add_member(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("added member"))

    # Check they are a member
    topology.logcap.flush()
    g_args.cn = "testgroup"
    members(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("uid=testuser,ou=people,dc=example,dc=com"))

    # Remove them from the group
    topology.logcap.flush()
    g_args.cn = "testgroup"
    g_args.dn = "uid=testuser,ou=people,dc=example,dc=com"
    remove_member(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("removed member"))

    # Check they are not a member
    topology.logcap.flush()
    g_args.cn = "testgroup"
    members(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args)
    assert(topology.logcap.contains("No members to display"))

    # Delete the group
    topology.logcap.flush()
    g_args.dn = "cn=testgroup,ou=groups,dc=example,dc=com"
    delete(topology.standalone, DEFAULT_SUFFIX, topology.logcap.log, g_args, warn=False)
    assert(topology.logcap.contains("Successfully deleted cn=testgroup,ou=groups,dc=example,dc=com"))



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ----


from enum import Enum
import ldap
from lib389._mapped_object import DSLdapObject, DSLdapObjects
from lib389.cos import CosTemplates, CosClassicDefinitions
from lib389.mappingTree import MappingTrees
from lib389.idm.nscontainer import nsContainers

MUST_ATTRIBUTES = [
    'cn',
]
MUST_ATTRIBUTES_NESTED = [
    'cn',
    'nsroledn'
]
MUST_ATTRIBUTES_FILTERED = [
    'cn',
    'nsrolefilter'
]
RDN = 'cn'


class RoleState(Enum):
    ACTIVATED = "activated"
    DIRECTLY_LOCKED = "directly locked through nsDisabledRole"
    INDIRECTLY_LOCKED = "indirectly locked through a Role"
    PROBABLY_ACTIVATED = '''probably activated or nsDisabledRole setup and its CoS entries are not
in a valid state or there is no access to the settings.'''

    def describe(self, role_dn=None):
        if self.name == "INDIRECTLY_LOCKED" and role_dn is not None:
            return f'{self.value} - {role_dn}'
        else:
            return f'{self.value}'


class Role(DSLdapObject):
    """A single instance of Role entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(Role, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        self._create_objectclasses = [
            'top',
            'LDAPsubentry',
            'nsRoleDefinition',
        ]
        self._protected = False

    def _format_status_message(self, message, role_dn=None):
        return {"state": message, "role_dn": role_dn}

    def status(self):
        """Check if role is locked in nsDisabledRole (directly or indirectly)

        :returns: a dict
        """

        inst = self._instance
        disabled_roles = {}
        try:
            mapping_trees = MappingTrees(inst)
            root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)
            roles = Roles(inst, root_suffix)
            disabled_roles = roles.get_disabled_roles()
            nested_roles = NestedRoles(inst, root_suffix)
            disabled_role = nested_roles.get("nsDisabledRole")
            inact_containers = nsContainers(inst, basedn=root_suffix)
            inact_container = inact_containers.get('nsAccountInactivationTmp')

            cos_templates = CosTemplates(inst, inact_container.dn)
            cos_template = cos_templates.get(f'{disabled_role.dn}')
            cos_template.present('cosPriority', '1')
            cos_template.present('nsAccountLock', 'true')

            cos_classic_defs = CosClassicDefinitions(inst, root_suffix)
            cos_classic_def = cos_classic_defs.get('nsAccountInactivation_cos')
            cos_classic_def.present('cosAttribute', 'nsAccountLock operational')
            cos_classic_def.present('cosTemplateDn', inact_container.dn)
            cos_classic_def.present('cosSpecifier', 'nsRole')
        except ldap.NO_SUCH_OBJECT:
            return self._format_status_message(RoleState.PROBABLY_ACTIVATED)

        for role, parent in disabled_roles.items():
            if str.lower(self.dn) == str.lower(role.dn):
                if parent is None:
                    return self._format_status_message(RoleState.DIRECTLY_LOCKED)
                else:
                    return self._format_status_message(RoleState.INDIRECTLY_LOCKED, parent)

        return self._format_status_message(RoleState.ACTIVATED)

    def lock(self):
        """Set the entry dn to nsDisabledRole and ensure it exists"""

        current_status = self.status()
        if current_status["state"] == RoleState.DIRECTLY_LOCKED:
            raise ValueError(f"Role is already {current_status['state'].describe()}")

        inst = self._instance

        mapping_trees = MappingTrees(inst)
        root_suffix = ""
        root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)

        if root_suffix:
            managed_roles = ManagedRoles(inst, root_suffix)
            managed_role = managed_roles.ensure_state(properties={"cn": "nsManagedDisabledRole"})
            nested_roles = NestedRoles(inst, root_suffix)
            try:
                disabled_role = nested_roles.get("nsDisabledRole")
            except ldap.NO_SUCH_OBJECT:
                # We don't use "ensure_state" because we want to preserve the existing attributes
                disabled_role = nested_roles.create(properties={"cn": "nsDisabledRole",
                                                                "nsroledn": managed_role.dn})
            disabled_role.add("nsRoleDN", self.dn)

            inact_containers = nsContainers(inst, basedn=root_suffix)
            inact_container = inact_containers.ensure_state(properties={'cn': 'nsAccountInactivationTmp'})

            cos_templates = CosTemplates(inst, inact_container.dn)
            cos_templates.ensure_state(properties={'cosPriority': '1',
                                                   'nsAccountLock': 'true',
                                                   'cn': f'{disabled_role.dn}'})

            cos_classic_defs = CosClassicDefinitions(inst, root_suffix)
            cos_classic_defs.ensure_state(properties={'cosAttribute': 'nsAccountLock operational',
                                                      'cosSpecifier': 'nsRole',
                                                      'cosTemplateDn': inact_container.dn,
                                                      'cn': 'nsAccountInactivation_cos'})

    def unlock(self):
        """Remove the entry dn from nsDisabledRole if it exists"""

        inst = self._instance
        current_status = self.status()
        if current_status["state"] == RoleState.ACTIVATED:
            raise ValueError("Role is already active")

        mapping_trees = MappingTrees(inst)
        root_suffix = mapping_trees.get_root_suffix_by_entry(self.dn)
        roles = NestedRoles(inst, root_suffix)
        try:
            disabled_role = roles.get("nsDisabledRole")
            # Still we want to ensure that it is not locked directly too
            disabled_role.ensure_removed("nsRoleDN", self.dn)
        except ldap.NO_SUCH_OBJECT:
            pass

        # Notify if it's locked indirectly
        if current_status["state"] == RoleState.INDIRECTLY_LOCKED:
            raise ValueError(f"Role is {current_status['state'].describe(current_status['role_dn'])}. Please, deal with it separately")


class Roles(DSLdapObjects):
    """DSLdapObjects that represents all Roles entries

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    """

    def __init__(self, instance, basedn):
        super(Roles, self).__init__(instance)
        self._objectclasses = [
            'top',
            'LDAPsubentry',
            'nsRoleDefinition',
        ]
        self._filterattrs = ['cn']
        self._basedn = basedn
        self._childobject = Role

    def get_with_type(self, selector=[], dn=None):
        """Get the correct role type

        :param dn: DN of wanted entry
        :type dn: str
        :param selector: An additional filter to search for, i.e. 'backend_name'. The attributes
                         selected are based on object type, ie user will search for uid and cn.
        :type dn: str

        :returns: FilteredRole, ManagedRole or NestedRole
        """

        ROLE_OBJECTCLASSES = {FilteredRole: ['nscomplexroledefinition',
                                             'nsfilteredroledefinition'],
                              ManagedRole: ['nssimpleroledefinition',
                                            'nsmanagedroledefinition'],
                              NestedRole: ['nscomplexroledefinition',
                                           'nsnestedroledefinition']}
        entry = self.get(selector=selector, dn=dn, json=False)
        entry_objectclasses = entry.get_attr_vals_utf8_l("objectClass")
        role_found = False
        for role, objectclasses in ROLE_OBJECTCLASSES.items():
            role_found = all(oc in entry_objectclasses for oc in objectclasses)
            if role_found:
                return role(self._instance, entry.dn)
        if not role_found:
            raise ldap.NO_SUCH_OBJECT("Role definition was not found")

    def get_disabled_roles(self):
        """Get disabled roles that are usually defined in the cn=nsDisabledRole,ROOT_SUFFIX

        :returns: A dict {role: its_parent, }
        """

        disabled_role = self.get("nsDisabledRole")
        roles_inactive = {}
        result = {}

        # Do this on 0 level of nestedness
        for role_dn in disabled_role.get_attr_vals_utf8_l("nsRoleDN"):
            roles_inactive[role_dn] = None

        # We go through the list and check if the role is Nested and
        # then add its 'nsrole' attributes to the processing list
        while roles_inactive.items():
            processing_role_dn, parent = roles_inactive.popitem()
            # Check if already seen the role and skip it then
            if processing_role_dn in result.keys():
                continue

            processing_role = self.get_with_type(dn=processing_role_dn)
            if isinstance(processing_role, NestedRole):
                for role_dn in processing_role.get_attr_vals_utf8_l("nsRoleDN"):
                    # We don't need to process children which are already present in the list
                    if role_dn in result.keys() or role_dn in roles_inactive.keys():
                        continue
                    # We are deeper - return its children to the processing and assign the original parent
                    if parent in [role.dn for role in result.keys()]:
                        roles_inactive[role_dn] = parent
                    else:
                        roles_inactive[role_dn] = processing_role_dn
            # Set the processed role to list
            result[processing_role] = parent

        return result


class FilteredRole(Role):
    """A single instance of FilteredRole entry to create FilteredRole role

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(FilteredRole, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        self._create_objectclasses = ['nsComplexRoleDefinition', 'nsFilteredRoleDefinition']
        self._protected = False


class FilteredRoles(Roles):
    """DSLdapObjects that represents all filtered role entries

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    """

    def __init__(self, instance, basedn):
        super(FilteredRoles, self).__init__(instance, basedn)
        self._objectclasses = ['LDAPsubentry', 'nsComplexRoleDefinition', 'nsFilteredRoleDefinition']
        self._must_attributes = MUST_ATTRIBUTES_FILTERED
        self._filterattrs = ['cn']
        self._basedn = basedn
        self._childobject = FilteredRole


class ManagedRole(Role):
    """A single instance of Managed Role entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(ManagedRole, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        self._create_objectclasses = ['nsSimpleRoleDefinition', 'nsManagedRoleDefinition']
        self._protected = False


class ManagedRoles(Roles):
    """DSLdapObjects that represents all Managed Roles entries

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined wit basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn):
        super(ManagedRoles, self).__init__(instance, basedn)
        self._objectclasses = ['LDAPsubentry', 'nsSimpleRoleDefinition', 'nsManagedRoleDefinition']
        self._filterattrs = ['cn']
        self._basedn = basedn
        self._childobject = ManagedRole


class NestedRole(Role):
    """A single instance of Nested Role entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(NestedRole, self).__init__(instance, dn)
        self._must_attributes = MUST_ATTRIBUTES_NESTED
        self._rdn_attribute = RDN
        self._create_objectclasses = ['nsComplexRoleDefinition', 'nsNestedRoleDefinition']
        self._protected = False


class NestedRoles(Roles):
    """DSLdapObjects that represents all NestedRoles entries in suffix.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined wit basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn):
        super(NestedRoles, self).__init__(instance, basedn)
        self._objectclasses = ['LDAPsubentry', 'nsComplexRoleDefinition', 'nsNestedRoleDefinition']
        self._filterattrs = ['cn']
        self._basedn = basedn
        self._childobject = NestedRole



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import copy
import os
import base64
import time
import fnmatch
from struct import pack, unpack
from datetime import timedelta
from stat import ST_MODE
# from lib389.utils import print_nice_time
from lib389.paths import Paths
from lib389.utils import ensure_str
from lib389._mapped_object_lint import DSLint
from lib389.lint import (
    DSPERMLE0001,
    DSPERMLE0002,
    DSSKEWLE0001,
    DSSKEWLE0002,
    DSSKEWLE0003,
    DSSKEWLE0004
)


class DSEldif(DSLint):
    """A class for working with dse.ldif file

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance, serverid=None, path=None):
        self._instance = instance
        self._contents = []

        if path:
            self.path = path
        elif serverid:
            # Get the dse.ldif from the instance name
            prefix = os.environ.get('PREFIX', ""),
            if serverid.startswith("slapd-"):
                serverid = serverid.replace("slapd-", "", 1)
            self.path = "{}/etc/dirsrv/slapd-{}/dse.ldif".format(prefix[0], serverid)
        else:
            ds_paths = Paths(self._instance.serverid, self._instance)
            self.path = os.path.join(ds_paths.config_dir, 'dse.ldif')

        with open(self.path, 'r') as file_dse:
            processed_line = ""
            for line in file_dse.readlines():
                if not line.startswith(' '):
                    if processed_line:
                        self._contents.append(processed_line)
                    if line.startswith('dn:'):
                        processed_line = line.lower()
                    else:
                        processed_line = line
                else:
                    processed_line = processed_line[:-1] + line[1:]

    @classmethod
    def lint_uid(cls):
        return 'dseldif'

    def _lint_nsstate(self):
        """
        Check the nsState attribute, which contains the CSN generator time
        diffs, for excessive replication time skew
        """
        ignoring_skew = False
        skew_high = 86400  # 1 day
        skew_medium = 43200  # 12 hours
        skew_low = 21600  # 6 hours

        ignore_skew = self.get("cn=config", "nsslapd-ignore-time-skew")
        if ignore_skew is not None and ignore_skew[0].lower() == "on":
            # If we are ignoring time skew only report a warning if the skew
            # is significant
            ignoring_skew = True
            skew_high = 86400 * 365  # Report a warning for skew over a year
            skew_medium = 99999999999
            skew_low = 99999999999

        suffixes = self.readNsState()
        for suffix in suffixes:
            # Check the local offset first
            report = None
            skew = abs(int(suffix['time_skew']))
            if skew >= skew_high:
                if ignoring_skew:
                    # Ignoring skew, but it's too excessive not to report it
                    report = copy.deepcopy(DSSKEWLE0004)
                else:
                    # 24 hours of skew - replication will break
                    report = copy.deepcopy(DSSKEWLE0003)
            elif skew >= skew_medium:
                # 12 hours
                report = copy.deepcopy(DSSKEWLE0002)
            elif skew >= skew_low:
                # 6 hours
                report = copy.deepcopy(DSSKEWLE0001)
            if report is not None:
                report['items'].append(suffix['suffix'])
                report['items'].append('Time Skew')
                report['items'].append('Skew: ' + suffix['time_skew_str'])
                report['fix'] = report['fix'].replace('YOUR_INSTANCE',
                                                      self._instance.serverid)
                report['check'] = 'dseldif:nsstate'
                yield report

    def _update(self):
        """Update the dse.ldif with a new contents"""

        with open(self.path, "w") as file_dse:
            file_dse.write("".join(self._contents))

    def globalSubstitute(self, strfrom, strto):
        for i in range(0, len(self._contents)-1):
            self._contents[i] = self._contents[i].replace(strfrom, strto)
        self._update()

    def _find_attr(self, entry_dn, attr, lower=False):
        """Find all attribute values and indexes under a given entry

        Returns entry dn index and attribute data dict:
        relative attribute indexes and the attribute value

        :param lower: Use case-insensitive matching for attribute name
        :type lower: boolean
        """

        entry_dn_i = self._contents.index("dn: {}\n".format(entry_dn.lower()))
        attr_data = {}

        # Find where the entry ends
        try:
            dn_end_i = self._contents[entry_dn_i:].index("\n")
        except ValueError:
            # We are in the end of the list
            dn_end_i = len(self._contents)

        entry_slice = self._contents[entry_dn_i:entry_dn_i + dn_end_i]

        # Find the attribute
        for line in entry_slice:
            if lower:
                match = line.lower().startswith("{}:".format(attr.lower()))
            else:
                match = line.startswith("{}:".format(attr))
            if match:
                attr_value = line.split(" ", 1)[1][:-1]
                attr_data.update({entry_slice.index(line): attr_value})

        if not attr_data:
            raise ValueError("Attribute {} wasn't found under dn: {}".format(attr, entry_dn.lower()))

        return entry_dn_i, attr_data

    def get(self, entry_dn, attr, single=False, lower=False):
        """Return attribute values under a given entry

        :param entry_dn: a DN of entry we want to get attribute from
        :type entry_dn: str
        :param attr: an attribute name
        :type attr: str
        :param single: Return a single value instead of a list
        :type single: boolean
        :param lower: Use case-insensitive matching for attribute name
        :type lower: boolean
        """

        try:
            _, attr_data = self._find_attr(entry_dn, attr, lower=lower)
        except ValueError:
            return None

        vals = list(attr_data.values())
        if single:
            return vals[0] if len(vals) > 0 else None
        return vals

    def get_entry_attrs(self, entry_dn):
        """Return all attributes for one entry as ``{attr_lower: [str, ...]}`` from ``dse.ldif``.

        Used when the directory server is not running but the instance ``dse.ldif`` is readable.
        """
        entry_dn = ensure_str(entry_dn).lower()
        dn_line = "dn: {}\n".format(entry_dn)
        try:
            entry_dn_i = self._contents.index(dn_line)
        except ValueError:
            return {}

        end_i = len(self._contents)
        for j in range(entry_dn_i + 1, len(self._contents)):
            if self._contents[j].startswith("dn: "):
                end_i = j
                break

        attrs = {}
        for line in self._contents[entry_dn_i + 1 : end_i]:
            line = line.rstrip("\n")
            if not line or line.startswith(" "):
                continue
            name, sep, rest = line.partition(":")
            if not sep:
                continue
            name = name.strip()
            rest = rest.lstrip()
            aname = name.lower()
            if rest.startswith(":"):
                try:
                    val = base64.b64decode(rest[1:].strip()).decode("utf-8")
                except Exception:
                    val = rest[1:].strip()
            else:
                val = rest
            if aname not in attrs:
                attrs[aname] = []
            attrs[aname].append(val)
        return attrs

    def get_indexes(self, backend):
        """Return a list of backend indexes

        :param backend: a backend to get the indexes of
        """
        indexes = []
        for entry in self._contents:
            if fnmatch.fnmatch(entry, "*,cn=index,cn={}*".format(backend.lower())):
                start = entry.find("cn=")
                end = entry.find(",")
                indexes.append(entry[start+len('cn='):end])

        return indexes

    def get_backends(self):
        """Return a list of backend names from DSE.

        Returns backend names preserving their original case, as the
        database directory names on disk use the original case.

        Note: DSEldif lowercases DN lines, so we read the 'cn' attribute
        from each entry to get the original case.

        :returns: List of backend names
        """
        backends = []
        excluded = ("config", "monitor", "index", "encrypted attributes")

        for entry in self._contents:
            if (entry.startswith("dn: cn=") and
                    ",cn=ldbm database,cn=plugins,cn=config" in entry):
                parts = entry.split(",")
                if len(parts) > 1:
                    cn_lower = parts[0].replace("dn: cn=", "")
                    if cn_lower not in excluded:
                        dn = entry.strip()[4:].strip()
                        try:
                            suffix = self.get(dn, "nsslapd-suffix")
                            if suffix:
                                cn_values = self.get(dn, "cn")
                                if cn_values:
                                    backends.append(cn_values[0])
                        except (ValueError, IndexError):
                            pass

        return list(set(backends))

    def get_mapping_trees(self):
        """Return a list of mapping tree names from DSE.
        """
        mapping_trees = []
        for entry in self._contents:
            if fnmatch.fnmatch(entry, "dn: cn=*,cn=mapping tree,cn=config*"):
                mapping_trees.append(entry.strip())
        return mapping_trees

    def get_replicas(self):
        """Return a list of replica DN's from DSE.

        :returns: List of replica DN's
        """
        replicas = []
        for entry in self._contents:
            if fnmatch.fnmatch(entry, "dn: cn=replica,cn=*,cn=mapping tree,cn=config*"):
                replicas.append(entry.strip()[4:].strip())

        return replicas

    def suffix_replicated(self, suffix):
        """
        Check if the suffix is replicated in the DSE.
        """
        for entry in self._contents:
            if fnmatch.fnmatch(entry, "dn: *cn=replica,cn=*,cn=mapping tree,cn=config*"):
                dn = entry.replace("dn: ", "").strip()
                vals = self.get(dn, "nsDS5ReplicaRoot")
                if vals is not None and vals[0].lower() == suffix.lower():
                    return True

        return False


    def add_entry(self, entry):
        """Add a new entry

        :param entry: the entry to add in ldif format
        :type value: str list
        """

        if self._contents[-1] != "\n":
            self._contents.append("\n")
        self._contents.extend(entry)
        self._update()


    def add(self, entry_dn, attr, value):
        """Add an attribute under a given entry

        :param entry_dn: a DN of entry we want to edit
        :type entry_dn: str
        :param attr: an attribute name
        :type attr: str
        :param value: an attribute value
        :type value: str
        """

        entry_dn_i = self._contents.index("dn: {}\n".format(entry_dn.lower()))
        self._contents.insert(entry_dn_i+1, "{}: {}\n".format(attr, value))
        self._update()

    def rename(self, entry_dn, new_dn, del_old_rdn=True):
        """Rename an entry
        :param entry_dn: a DN of entry we want to rename
        :type entry_dn: str
        :param new_dn: The new DN
        :type new_dn: str
        :param del_old_rdn: Delete the old RND attribute from the entry
        :type del_old_rdn: boolean
        """

        # old rdn
        old_rdn = entry_dn.split(',', 1)[0]
        rdn_attr = old_rdn.split('=')[0]

        # new rdn
        new_rdn = new_dn.split(',', 1)[0]
        new_rdn_attr = new_rdn.split('=')[0]
        new_rdn_val = new_rdn.split('=')[1]

        # Handle the rdn attribute
        if del_old_rdn:
           self.delete(entry_dn, rdn_attr)
        self.add(entry_dn, new_rdn_attr, new_rdn_val)

        # Rename the entry
        entry_dn_i = self._contents.index("dn: {}\n".format(entry_dn.lower()))
        self._contents[entry_dn_i] = f"dn: {new_dn}\n"
        self._update()

    def delete_dn(self, entry_dn):
        """Delete the whole entry by DN

        :param entry_dn: a DN of an entry we want to delete
        :type entry_dn: str
        """

        entry_dn_i = self._contents.index("dn: {}\n".format(entry_dn.lower()))

        # Find where the entry ends
        try:
            dn_end_i = self._contents[entry_dn_i:].index("\n")
        except ValueError:
            # We are in the end of the list
            dn_end_i = len(self._contents)

        del self._contents[entry_dn_i:entry_dn_i+dn_end_i+1]
        self._update()

    def delete(self, entry_dn, attr, value=None):
        """Delete attributes under a given entry

        :param entry_dn: a DN of entry we want to edit
        :type entry_dn: str
        :param attr: an attribute name
        :type attr: str
        :param value: an attribute value
        :type value: str
        """

        entry_dn_i, attr_data = self._find_attr(entry_dn, attr)

        if value is not None:
            for attr_i, attr_value in attr_data.items():
                if attr_value == value:
                    del self._contents[entry_dn_i + attr_i]
        else:
            for attr_i in sorted(attr_data.keys(), reverse=True):
                del self._contents[entry_dn_i + attr_i]
        self._update()

    def replace(self, entry_dn, attr, value):
        """Replace attribute values with a new one under a given entry

        :param entry_dn: a DN of entry we want to edit
        :type entry_dn: str
        :param attr: an attribute name
        :type attr: str
        :param value: an attribute value
        :type value: str
        """

        try:
            self.delete(entry_dn, attr)
        except ValueError as e:
            self._instance.log.debug("During replace operation: {}".format(e))
        self.add(entry_dn, attr, value)
        self._update()

    # Read NsState helper functions
    def _flipend(self, end):
        if end == '<':
            return '>'
        if end == '>':
            return '<'

    def _getGenState(self, dn, replica_suffix, nsstate, flip):
        """Return a dict ofall the nsState properties
        """
        from lib389.utils import print_nice_time
        if pack('<h', 1) == pack('=h',1):
            endian = "Little Endian"
            end = '<'
            if flip:
                end = self._flipend(end)
        elif pack('>h', 1) == pack('=h',1):
            endian = "Big Endian"
            end = '>'
            if flip:
                end = self._flipend(end)
        else:
            raise ValueError("Unknown endian, unable to proceed")

        thelen = len(nsstate)
        if thelen <= 20:
            pad = 2 # padding for short H values
            timefmt = 'I' # timevals are unsigned 32-bit int
        else:
            pad = 6 # padding for short H values
            timefmt = 'Q' # timevals are unsigned 64-bit int

        base_fmtstr = "H%dx3%sH%dx" % (pad, timefmt, pad)
        fmtstr = end + base_fmtstr
        (rid, sampled_time, local_offset, remote_offset, seq_num) = unpack(fmtstr, nsstate)
        now = int(time.time())
        tdiff = now-sampled_time
        wrongendian = False
        try:
            tdelta = timedelta(seconds=tdiff)
            wrongendian = tdelta.days > 10*365
        except OverflowError: # int overflow
            wrongendian = True

        # if the sampled time is more than 20 years off, this is
        # probably the wrong endianness
        if wrongendian:
            end = self._flipend(end)
            fmtstr = end + base_fmtstr
            (rid, sampled_time, local_offset, remote_offset, seq_num) = unpack(fmtstr, nsstate)
            tdiff = now-sampled_time
            tdelta = timedelta(seconds=tdiff)

        return {
            'dn': dn,
            'suffix': replica_suffix,
            'endian': endian,
            'rid': str(rid),
            'gen_time': str(sampled_time),
            'gencsn': "%08x%04x%04x0000" % (sampled_time, seq_num, rid),
            'gen_time_str': time.ctime(sampled_time),
            'local_offset': str(local_offset),
            'local_offset_str': print_nice_time(local_offset),
            'remote_offset': str(remote_offset),
            'remote_offset_str': print_nice_time(remote_offset),
            'time_skew': str(local_offset + remote_offset),
            'time_skew_str': print_nice_time(local_offset + remote_offset),
            'seq_num': str(seq_num),
            'sys_time': str(time.ctime(now)),
            'diff_secs': str(tdiff),
            'diff_days_secs': "%d:%d" % (tdelta.days, tdelta.seconds),
        }

    def readNsState(self, suffix=None, flip=False):
        """Look for the nsState attribute in replication configuration entries,
        then decode the base64 value and  provide a dict of all stats it
        contains

        :param suffix: specific suffix to read nsState from
        :type suffix: str
        """
        found_replica = False
        found_suffix = False
        replica_suffix = ""
        nsstate = ""
        states = []

        for line in self._contents:
            if line.startswith("dn: "):
                dn = line[4:].strip()
                if dn.startswith("cn=replica"):
                    found_replica = True
                else:
                    found_replica = False
            else:
                if line.lower().startswith("nsstate:: ") and found_replica:
                    b64val = line[10:].strip()
                    nsstate = base64.decodebytes(b64val.encode())
                elif line.lower().startswith("nsds5replicaroot"):
                    found_suffix = True
                    replica_suffix = line.lower().split(':')[1].strip()

            if found_replica and found_suffix and nsstate != "":
                # We have everything we need to proceed
                if suffix is not None and suffix == replica_suffix:
                    states.append(self._getGenState(dn, replica_suffix, nsstate, flip))
                    break
                else:
                    states.append(self._getGenState(dn, replica_suffix, nsstate, flip))
                    # reset flags for next round...
                    found_replica = False
                    found_suffix = False
                    replica_suffix = ""
                    nsstate = ""

        return states

    def _increaseTimeSkew(self, suffix, timeSkew):
        # Increase csngen state local_offset by timeSkew
        # Warning: instance must be stopped before calling this function
        assert (timeSkew >= 0)
        nsState = self.readNsState(suffix)[0]
        self._instance.log.debug(f'_increaseTimeSkew nsState is {nsState}')
        oldNsState = self.get(nsState['dn'], 'nsState', True)
        self._instance.log.debug(f'oldNsState is {oldNsState}')

        # Lets reencode the new nsState
        from lib389.utils import print_nice_time
        if pack('<h', 1) == pack('=h',1):
            end = '<'
        elif pack('>h', 1) == pack('=h',1):
            end = '>'
        else:
            raise ValueError("Unknown endian, unable to proceed")

        thelen = len(oldNsState)
        if thelen <= 20:
            pad = 2 # padding for short H values
            timefmt = 'I' # timevals are unsigned 32-bit int
        else:
            pad = 6 # padding for short H values
            timefmt = 'Q' # timevals are unsigned 64-bit int
        fmtstr = "%sH%dx3%sH%dx" % (end, pad, timefmt, pad)
        newNsState = base64.b64encode(pack(fmtstr, int(nsState['rid']),
           int(nsState['gen_time']), int(nsState['local_offset'])+timeSkew,
           int(nsState['remote_offset']), int(nsState['seq_num'])))
        newNsState = newNsState.decode('utf-8')
        self._instance.log.debug(f'newNsState is {newNsState}')
        # Lets replace the value.
        (entry_dn_i, attr_data) = self._find_attr(nsState['dn'], 'nsState')
        attr_i = next(iter(attr_data))
        self._contents[entry_dn_i + attr_i] = f"nsState:: {newNsState}"
        self._update()


class FSChecks(DSLint):
    """This is for the healthcheck feature, check commonly used system config files the
    server uses.  This is here for lack of a better place to add this class.
    """
    def __init__(self, dirsrv=None):
        self.dirsrv = dirsrv
        self._certdb = self.dirsrv.get_cert_dir()
        self.ds_files = [
            {
                'name': '/etc/resolv.conf',
                'perms': [644],
                'report': DSPERMLE0001
            },
            {
                'name': self._certdb + "/pin.txt",
                'perms': [400, 600],
                'report': DSPERMLE0002
            },
            {
                'name': self._certdb + "/pwdfile.txt",
                'perms': [400, 600],
                'report': DSPERMLE0002
            },
        ]

    @classmethod
    def lint_uid(cls):
        return 'fschecks'

    def _lint_file_perms(self):
        """Test file permissions are safe
        """
        for ds_file in self.ds_files:
            try:
                perms = int(oct(os.stat(ds_file['name'])[ST_MODE])[-3:])
                if perms not in ds_file['perms']:
                    perms = str(ds_file['perms'][0])
                    report = copy.deepcopy(ds_file['report'])
                    report['items'].append(ds_file['name'])
                    report['detail'] = report['detail'].replace('FILE', ds_file['name'])
                    report['detail'] = report['detail'].replace('PERMS', perms)
                    report['fix'] = report['fix'].replace('FILE', ds_file['name'])
                    report['fix'] = report['fix'].replace('PERMS', perms)
                    report['check'] = f'fschecks:file_perms'
                    yield report
            except FileNotFoundError:
                pass



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import logging
from typing import Optional
from lib389 import DirSrv
from lib389.nss_ssl import NssSsl
from lib389.dyncerts import DynamicCerts
from lib389.config import RSA

log = logging.getLogger(__name__)

class CertManager:
    """
    Certificate manager for 389 Directory Server.

    Automatically selects DynamicCerts backend if available via LDAPI,
    otherwise falls back to NSS DB.
    """

    def __init__(self, instance):
        """
        Initialise a CertManager object, which selects the appropriate
        certificate handler (DynamicCert or NSS).

        :param instance: DirSrv instance
        :raises ValueError: If instance is None
        """
        if not isinstance(instance, DirSrv):
            raise ValueError("A DirSrv instance is required")

        self.dirsrv = instance

        if self.dirsrv.status():
            self.cert_handler = DynamicCerts(instance=self.dirsrv)
        else:
            self.cert_handler = NssSsl(dirsrv=self.dirsrv)

        self.cert_handler_name = type(self.cert_handler).__name__

    def list_certs(self):
        """
        Return a list of all certificates exposed by the backend.

        :return: list of certificate dictionaries or tuples
        """
        return self.cert_handler.list_certs()

    def list_ca_certs(self):
        """
        Return a list of all CA certificates exposed by the backend.

        :return: list of CA certificate dictionaries or tuples
        """
        return self.cert_handler.list_ca_certs()

    def get_cert(self, nickname: str):
        """
        Get certificate details by nickname/CN.

        :param nickname: Certificate nickname
        :return: backend certificate object or None
        """
        return self.cert_handler.get_cert_details(nickname=nickname)

    def del_cert(self, nickname: str):
        """
        Delete a certificate by nickname.

        :param nickname: Certificate nickname
        :raises ValueError: If the nickname is empty.
        """
        if not nickname:
            raise ValueError("Certificate nickname cannot be empty")

        self.cert_handler.del_cert(nickname)

    def add_cert(
        self,
        cert_file: str,
        nickname: str,
        pkcs12_password: Optional[str] = None,
        primary: bool = False,
        ca: bool = False,
        force: bool = False
    ):
        """
        Add or replace a certificate.

        :param cert_file: Path to certificate file (PEM, DER, or PKCS#12)
        :param nickname: Certificate nickname/CN
        :param pkcs12_password: Password for PKCS#12, if any
        :param primary: Set as the server's primary SSL certificate
        :param ca: Whether this certificate is a CA certificate
        :param force: Force the addition of a certificate that cannot be verified
        :raises ValueError: If cert_file or nickname is invalid
        """
        if not os.path.isfile(cert_file):
            raise ValueError(f"Certificate file not found: {cert_file}")
        if not nickname:
            raise ValueError("Certificate nickname must not be empty")

        self.cert_handler.add_cert(
            nickname=nickname,
            cert_file=cert_file,
            pkcs12_password=pkcs12_password,
            ca=ca,
            force=force
        )

        if primary:
            try:
                RSA(self.dirsrv).set("nsSSLPersonalitySSL", nickname)
                log.info(f"Set certificate '{nickname}' as primary SSL certificate.")
            except Exception as e:
                log.error(f"Failed to set primary SSL cert '{nickname}': {e}")
                raise

    def add_ca_cert(self, cert_file: str, nickname: str, force: bool = False):
        """
        Add one or more CA certificates from a PEM bundle or single DER.

        :param cert_file: Path to certificate file (PEM, DER)
        :param nickname: Certificate nickname
        :raises ValueError: If file is missing
        """
        if not os.path.exists(cert_file):
            raise ValueError(f"Certificate file does not exist: {cert_file}")

        self.cert_handler.add_ca_cert(cert_file, nickname, force=force)

    def edit_cert_trust(self, nickname: str, trust_flags: str):
        """
        Edit trust flags on an existing certificate.

        :param nickname: Certificate nickname
        :param trust_flags: NSS style trust flag triplet, e.g. 'CT,,'
        """
        return self.cert_handler.edit_cert_trust(nickname=nickname, trust_flags=trust_flags)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

# Replacement of the dbgen.pl utility

from lib389.utils import (ensure_str, pseudolocalize)
import random
import os
import pwd
import grp

global node_count

DBGEN_POSITIONS = [
"Accountant",
"Admin",
"Architect",
"Assistant",
"Artist",
"Consultant",
"Czar",
"Dictator",
"Director",
"Diva",
"Dreamer",
"Evangelist",
"Engineer",
"Figurehead",
"Fellow",
"Grunt",
"Guru",
"Janitor",
"Madonna",
"Manager",
"Pinhead",
"President",
"Punk",
"Sales Rep",
"Stooge",
"Visionary",
"Vice President",
"Writer",
"Warrior",
"Yahoo"
]

DBGEN_TITLE_LEVELS = [
"Senior",
"Principal",
"Associate",
"Junior",
"Chief",
"Supreme",
"Elite"
]

DBGEN_LOCATIONS = [
"Mountain View", "Redmond", "Redwood Shores", "Armonk",
"Cambridge", "Santa Clara", "Sunnyvale", "Alameda",
"Cupertino", "Menlo Park", "Palo Alto", "Orem",
"San Jose", "San Francisco", "Milpitas", "Hartford", "Windsor",
"Boston", "New York", "Detroit", "Dallas", "Denver", "Brisbane",
]

DBGEN_OUS = [
"accounting",
"product development",
"product testing",
"human resources",
"payroll",
"people",
"groups",
]

DBGEN_TEMPLATE = """dn: {DN}{CHANGETYPE}
objectClass: top
objectClass: person
objectClass: organizationalPerson
objectClass: inetOrgPerson
objectclass: inetUser
objectclass: posixAccount
cn: {CN}
sn: {LAST}
uid: {UID}
givenName: {FIRST}
description: 2;7613;CN=Red Hat CS 71GA Demo,O=Red Hat CS 71GA Demo,C=US;CN=RHCS Agent - admin01,UID=admin01,O=redhat,C=US [1] This is {FIRST} {LAST}'s description.
userPassword: {UID}
departmentNumber: 1230
employeeType: Manager
homePhone: +1 303 937-6482
initials: {INITIALS}
telephoneNumber: +1 303 573-9570
facsimileTelephoneNumber: +1 415 408-8176
mobile: +1 818 618-1671
pager: +1 804 339-6298
roomNumber: 5164
carLicense: 21SJJAG
l: {LOCATION}
ou: {OU}
mail: {UID}@example.com
mail: {UIDNUMBER}@example.com
uidNumber: {UIDNUMBER}
gidNumber: {UIDNUMBER}
homeDirectory: /home/{UID}
postalAddress: 518, Dept #851, Room#{OU}
title: {TITLE}
usercertificate;binary:: MIIBvjCCASegAwIBAgIBAjANBgkqhkiG9w0BAQQFADAnMQ8wDQYD
 VQQDEwZjb25maWcxFDASBgNVBAMTC01NUiBDQSBDZXJ0MB4XDTAxMDQwNTE1NTEwNloXDTExMDcw
 NTE1NTEwNlowIzELMAkGA1UEChMCZnIxFDASBgNVBAMTC01NUiBTMSBDZXJ0MIGfMA0GCSqGSIb3
 DQEBAQUAA4GNADCBiQKBgQDNlmsKEaPD+o3mAUwmW4E40MPs7aiui1YhorST3KzVngMqe5PbObUH
 MeJN7CLbq9SjXvdB3y2AoVl/s5UkgGz8krmJ8ELfUCU95AQls321RwBdLRjioiQ3MGJiFjxwYRIV
 j1CUTuX1y8dC7BWvZ1/EB0yv0QDtp2oVMUeoK9/9sQIDAQABMA0GCSqGSIb3DQEBBAUAA4GBADev
 hxY6QyDMK3Mnr7vLGe/HWEZCObF+qEo2zWScGH0Q+dAmhkCCkNeHJoqGN4NWjTdnBcGaAr5Y85k1
 o/vOAMBsZePbYx4SrywL0b/OkOmQX+mQwieC2IQzvaBRyaNMh309vrF4w5kExReKfjR/gXpHiWQz
 GSxC5LeQG4k3IP34

"""

DBGEN_OU_TEMPLATE = """dn: ou={OU},{SUFFIX}
objectClass: top
objectClass: organizationalUnit
ou: {OU}

"""

RANDOM_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqurstuvwxyz0123456789_#@%&()?~$^`~*-=+{}|"\'.,<>'


def finalize_ldif_file(instance, ldif_file):
    # Make the file owned by dirsrv
    os.chmod(ldif_file, 0o644)
    if os.getuid() == 0:
        # root user - chown the ldif to the server user
        userid = ensure_str(instance.userid)
        uid = pwd.getpwnam(userid).pw_uid
        gid = grp.getgrnam(userid).gr_gid
        os.chown(ldif_file, uid, gid)


def get_index(idx, numUsers):
    # Get ldclt style entry "0" padded index number
    zeroLen = len(str(numUsers)) - len(str(idx))
    index = '0' * zeroLen
    index = index + str(idx)
    return index


def get_node(suffix):
    # Build a node/container entry based on the suffix DN
    rdn_attr = suffix.split('=')[0].lower()
    rdn_attr_val = suffix.split('=')[1].split(',')[0]
    if rdn_attr == 'c':
        oc = 'country'
    elif rdn_attr == 'cn':
        oc = 'nscontainer'
    elif rdn_attr == 'dc':
        oc = 'domain'
    elif rdn_attr == 'o':
        oc = 'organization'
    elif rdn_attr == 'ou':
        oc = 'organizationalunit'
    else:
        # Unsupported rdn
        raise ValueError("Suffix RDN '{}' in '{}' is not supported.  Supported RDN's are: 'c', 'cn', 'dc', 'o', and 'ou'".format(rdn_attr, suffix))

    return f"""dn: {suffix}
objectClass: top
objectClass: {oc}
{rdn_attr}: {rdn_attr_val}
aci: (target=ldap:///{suffix})(targetattr=*)(version 3.0; acl "Self Write"; allow(write) userdn = "ldap:///self";)
aci: (target=ldap:///{suffix})(targetattr=*)(version 3.0; acl "Directory Admin Group"; allow(write) groupdn = "ldap:///cn=Directory Administrators,ou=Groups,{suffix}";)
aci: (target=ldap:///{suffix})(targetattr=*)(version 3.0; acl "Anonymous Access"; allow(read, search, compare) userdn = "ldap:///anyone";)

"""


def randomPick(values):
    # Return a randomly selected value from the provided list of values
    val_count = len(values)
    val_count -= 1
    idx = random.randint(0, val_count)
    return values[idx].lstrip()


def write_generic_user(LDIF, index, number, parent, name="user", changetype="", pseudol10n=False):
    uid_val = name + get_index(index, number)
    ou = random.choice(DBGEN_OUS)
    first = uid_val
    last = uid_val[::-1]  # reverse name
    cn = f"{first} {last}"
    initials = "%s. %s" % (first[0], last[0])
    l = random.choice(DBGEN_LOCATIONS)
    title = "%s %s" % (random.choice(DBGEN_TITLE_LEVELS), random.choice(DBGEN_POSITIONS))
    if pseudol10n:
        ou = pseudolocalize(ou)
        first = pseudolocalize(first)
        last = pseudolocalize(last)
        initials = pseudolocalize(initials)
        l = pseudolocalize(l)
        title = pseudolocalize(title)
    dn = f"uid={uid_val},{parent}"

    LDIF.write(DBGEN_TEMPLATE.format(
        DN=dn,
        CHANGETYPE=changetype,
        UID=uid_val,
        UIDNUMBER=index,
        FIRST=first,
        LAST=last,
        CN=cn,
        INITIALS=initials,
        OU=ou,
        LOCATION=l,
        TITLE=title,
    ))
    return dn

def dbgen_users(instance, number, ldif_file, suffix, generic=False, entry_name="user", parent=None, startIdx=0, rdnCN=False, pseudol10n=False):
    """
    Generate an LDIF of randomly named entries
    """
    # Lets insure that integer parameters are not string
    number=int(number)
    startIdx=int(startIdx)
    familyname_file = os.path.join(instance.ds_paths.data_dir, 'dirsrv/data/dbgen-FamilyNames')
    givename_file = os.path.join(instance.ds_paths.data_dir, 'dirsrv/data/dbgen-GivenNames')
    familynames = []
    givennames = []
    with open(familyname_file, 'r') as f:
        familynames = [n.strip() for n in f]
    with open(givename_file, 'r') as f:
        givennames = [n.strip() for n in f]

    with open(ldif_file, 'w') as LDIF:
        LDIF.write(get_node(suffix))
        for ou in DBGEN_OUS:
            ou = pseudolocalize(ou) if pseudol10n else ou
            LDIF.write(DBGEN_OU_TEMPLATE.format(SUFFIX=suffix, OU=ou))

        if parent is not None:
            parent_rdn = parent.split(',')[0].split('=')[1]
            if parent_rdn.lower() not in DBGEN_OUS:
                LDIF.write(get_node(parent))

        for i in range(1, int(number) + 1):
            # Pick a random ou
            ou = random.choice(DBGEN_OUS)
            first = random.choice(givennames)
            last = random.choice(familynames)
            if generic:
                i += startIdx
                name = entry_name + get_index(i, number)
                uid = name
                cn = name
            else:
                first = random.choice(givennames)
                last = random.choice(familynames)
                uid = "%s%s%s" % (first[0], last, i)
                cn = f"{first} {last}"
            # How do we subscript from a generator?
            initials = "%s. %s" % (first[0], last[0])
            l = random.choice(DBGEN_LOCATIONS)
            title = "%s %s" % (random.choice(DBGEN_TITLE_LEVELS), random.choice(DBGEN_POSITIONS))
            if pseudol10n:
                ou = pseudolocalize(ou)
                first = pseudolocalize(first)
                last = pseudolocalize(last)
                initials = pseudolocalize(initials)
                l = pseudolocalize(l)
                title = pseudolocalize(title)

            if parent is None:
                parent = f"ou={ou},{suffix}"

            if rdnCN:
                # Not using "uid" so use "cn" instead
                dn = f"cn={cn},{parent}"
            else:
                dn = f"uid={uid},{parent}"

            LDIF.write(DBGEN_TEMPLATE.format(
                DN=dn,
                CHANGETYPE="",
                UID=uid,
                UIDNUMBER=i,
                FIRST=first,
                LAST=last,
                CN=cn,
                INITIALS=initials,
                OU=ou,
                LOCATION=l,
                TITLE=title,
            ))

    finalize_ldif_file(instance, ldif_file)


def dbgen_groups(instance, ldif_file, props):
    """
    Create static group(s) and the member entries

        props = {
            "name": STRING,
            "parent": DN,
            "suffix": DN,
            "number": ###,  --> number of groups to create
            "numMembers": ###,
            "createMembers": True/False  --> Create the member entries (default is True)
            "memberParent": DN
            "membershipAttr": ATTR
        }
    """
    with open(ldif_file, 'w') as LDIF:
        # Create the top node
        LDIF.write(get_node(props['suffix']))
        if props['parent'] is not None:
            if props['parent'] != props['suffix']:
                # Create the group container
                LDIF.write(get_node(props['parent']))
        else:
            props['parent'] = props['suffix']

        if props['memberParent'] is not None:
            if props['memberParent'] != props['suffix'] and props['memberParent'] != props['parent']:
                # Create the member/user container
                LDIF.write(get_node(props['memberParent']))
        else:
            props['memberParent'] = props['suffix']

        for idx in range(1, int(props['number']) + 1):
            # Build the member list and create the member entries
            group_member_list = []
            if props['createMembers']:
                for user_idx in range(1, int(props['numMembers']) + 1):
                    dn = write_generic_user(LDIF, user_idx, props['numMembers'], props['memberParent'], name=f"group_entry{idx}-")
                    group_member_list.append(dn)
            else:
                # Not creating the member entries, just build the DN list of members
                for user_idx in range(1, int(props['numMembers']) + 1):
                    name = "user" + get_index(user_idx, props['numMembers'])
                    dn = f"uid={name},{props['memberParent']}"
                    group_member_list.append(dn)

            if props['number'] == 0:
                # Only creating one group, do not add the idx to DN
                group_dn = f"dn: cn={props['name']},{props['parent']}\n"
                cn = f"cn={props['name']},{props['parent']}"
            else:
                group_dn = f"dn: cn={props['name']}-{idx},{props['parent']}\n"
                cn = f"cn={props['name']}-{idx},{props['parent']}"

            LDIF.write(group_dn)
            LDIF.write('objectclass: top\n')
            LDIF.write('objectclass: groupOfUniqueNames\n')
            LDIF.write('objectclass: groupOfNames\n')
            LDIF.write('objectclass: inetAdmin\n')
            LDIF.write(f'cn: {cn}\n')
            for dn in group_member_list:
                LDIF.write(f"{props['membershipAttr']}: {dn}\n")
            LDIF.write('\n')

    finalize_ldif_file(instance, ldif_file)


def dbgen_cos_def(instance, ldif_file, props):
    """
    Create a COS definition

        props = {
            "cosType": "classic", "pointer", "indirect",
            "defName": VAL,
            "defParent": VAL,
            "defCreateParent": True/False,
            "cosSpecifier": can be used for cosIndirectSpecifier
            "cosAttrs": [],
            "tmpName": DN (need to classic and pointer COS defs)
        }
    """

    if props['cosType'] == 'pointer':
        objectclass = 'objectclass: cosPointerDefinition\n'
    if props['cosType'] == 'indirect':
        objectclass = 'objectclass: cosIndirectDefinition\n'
    if props['cosType'] == 'classic':
        objectclass = 'objectclass: cosClassicDefinition\n'

    with open(ldif_file, 'w') as LDIF:
        # Create parent
        if props['defCreateParent']:
            LDIF.write(get_node(props['defParent']))

        #
        # Create definition
        #
        dn = (f"dn: cn={props['defName']},{props['defParent']}\n")
        LDIF.write(dn)
        LDIF.write('objectclass: top\n')
        LDIF.write('objectclass: cosSuperDefinition\n')
        LDIF.write(objectclass)
        LDIF.write('cn: ' + props['defName'] + "\n")

        if props['cosType'] == 'pointer' or props['cosType'] == 'classic':
            LDIF.write(f"cosTemplateDN: {props['tmpName']}\n")
        if props['cosType'] == 'indirect':
            LDIF.write(f"cosIndirectSpecifier: {props['cosSpecifier']}\n")
        elif props['cosType'] == "classic":
            LDIF.write(f"cosSpecifier: {props['cosSpecifier']}\n")
        for attr in props['cosAttrs']:
            # There can be multiple COS attributes
            LDIF.write(f"cosAttribute: {attr}\n")

    finalize_ldif_file(instance, ldif_file)


def dbgen_cos_template(instance, ldif_file, props):
    """
    Create a COS Template

        props = {
            "tmpName": VAL,
            "tmpParent": VAL,
            "tmpCreateParent": True/False,
            "cosPriority": ####
            "cosTmpAttrVal": Attr/val
        }
    """

    with open(ldif_file, 'w') as LDIF:
        # Create parent
        if props['tmpCreateParent']:
            LDIF.write(get_node(props['tmpParent']))

        # Create template
        dn = f"dn: cn={props['tmpName']},{props['tmpParent']}\n"
        LDIF.write(dn)
        LDIF.write('objectclass: top\n')
        LDIF.write('objectclass: extensibleObject\n')
        LDIF.write('objectclass: cosTemplate\n')
        LDIF.write(f"cn: {props['tmpName']}\n")
        if props['cosPriority'] is not None:
            LDIF.write(f"cosPriority: {props['cosPriority']}\n")
        pair = props['cosTmpAttrVal'].split(':')
        LDIF.write(f"{pair[0]}: {pair[1]}\n")

    finalize_ldif_file(instance, ldif_file)


def dbgen_role(instance, ldif_file, props):
    """
    Create a Role

        props = {
            role_type: "managed", "filter", pr "nested",
            role_name: NAME,
            parent: DN of parent,
            createParent: True/False,
            filter: FILTER,
            role_list: [DN, DN, ...]  # For nested role only
        }
    """

    if props['role_type'].lower() == 'managed':
        objectclasses = ('objectclass: nsSimpleRoleDefinition\n' +
                         'objectclass: nsManagedRoleDefinition\n')
    elif props['role_type'].lower() == 'filtered':
        objectclasses = ('objectclass: nsComplexRoleDefinition\n' +
                         'objectclass: nsFilteredRoleDefinition\n')
    elif props['role_type'].lower() == 'nested':
        objectclasses = ('objectclass: nsComplexRoleDefinition\n' +
                         'objectclass: nsNestedRoleDefinition\n')

    with open(ldif_file, 'w') as LDIF:
        # Create parent entry
        if props['createParent']:
            LDIF.write(get_node(props['parent']))

        dn = f"dn: cn={props['role_name']},{props['parent']}\n"
        LDIF.write(dn)
        LDIF.write('objectclass: top\n')
        LDIF.write('objectclass: LdapSubEntry\n')
        LDIF.write('objectclass: nsRoleDefinition\n')
        LDIF.write(objectclasses)
        LDIF.write(f"cn: {props['role_name']}\n")
        if props['role_type'] == 'nested':
            # Write out each role DN
            for value in props['role_list']:
                LDIF.write(f'nsRoleDN: {value}\n')
        elif props['role_type'] == 'filtered':
            LDIF.write(f"nsRoleFilter: {props['filter']}\n")

    finalize_ldif_file(instance, ldif_file)


def dbgen_mod_load(ldif_file, props):
    """
    Generate a "load" LDIF file that can be consumed by ldapmodify

        props = {
            "createUsers": True/False,
            "deleteUsers": True/False,
            "numUsers": ###,
            "parent": DN,  --> ou=people,dc=example,dc=com
            "createParent": True/False,
            "addUsers": ###,
            "delUsers": ###,
            'modrdnUsers": ###,
            "modUsers": ###,  --> number of entries to modify
            "random": True/False,
            "modAttrs": [ATTR, ATTR, ...]
        }
    """

    entry_dn_list = []  # List used to delete entries at the end of the LDIF
    if props['modAttrs'] is None:
        props['modAttrs'] = ['description', 'title']

    with open(ldif_file, 'w') as LDIF:
        if props['createParent']:
            # Create the container entry that the users will be add to
            LDIF.write(get_node(props['parent']))

        # Create entries
        for user_idx in range(1, props['numUsers'] + 1):
            if props['createUsers']:
                dn = write_generic_user(
                    LDIF, user_idx, props['numUsers'], props['parent'],
                    changetype="\nchangetype: add")
                entry_dn_list.append(dn)
            else:
                dn = f"uid=user{get_index(user_idx, props['numUsers'])},{props['parent']}"
                entry_dn_list.append(dn)

        # Set the types of operations and how many of them to perform
        addc = int(props['addUsers'])
        delc = int(props['delUsers'])
        modc = int(props['modUsers'])
        mrdnc = int(props['modrdnUsers'])
        total_ops = addc + delc + modc + mrdnc

        if props['random']:
            # Mix up the selected operations
            operations = ['add', 'mod', 'modrdn', 'delete']
            while total_ops != 0:
                op = randomPick(operations)
                if op == 'add':
                    if addc == 0:
                        # no more adds to do
                        operations.remove('add')
                        continue
                    dn = write_generic_user(
                        LDIF, addc, props['addUsers'], props['parent'],
                        name="addUser", changetype="\nchangetype: add")
                    entry_dn_list.append(dn)
                    addc -= 1
                elif op == 'mod':
                    if modc == 0:
                        # no more mods to do
                        operations.remove('mod')
                        continue
                    attr = randomPick(props['modAttrs'])
                    val = (''.join((random.choice(RANDOM_CHARS) for i in range(0, random.randint(10, 30)))))
                    LDIF.write(f"dn: uid=user{get_index(modc, props['numUsers'])},{props['parent']}\n")
                    LDIF.write("changetype: modify\n")
                    LDIF.write(f"replace: {attr}\n")
                    LDIF.write(f"{attr}: {val}\n")
                    LDIF.write("\n")
                    modc -= 1
                elif op == 'delete':
                    if delc == 0:
                        # no more deletes to do
                        operations.remove('delete')
                        continue

                    dn_val = f"uid=user{get_index(delc, props['numUsers'])},{props['parent']}"
                    LDIF.write(f"dn: {dn_val}\n")
                    LDIF.write("changetype: delete\n")
                    LDIF.write(" \n")
                    delc -= 1
                    if dn_val in entry_dn_list:
                        entry_dn_list.remove(dn_val)
                elif op == 'modrdn':
                    if mrdnc == 0:
                        # no more modrdns to do
                        operations.remove('modrdn')
                        continue
                    dn_val = f"uid=user{get_index(mrdnc, props['numUsers'])},{props['parent']}"
                    new_dn_val = f"cn=user{get_index(mrdnc, props['numUsers'])},{props['parent']}"
                    LDIF.write(f"dn: {dn_val}\n")
                    LDIF.write("changetype: modrdn\n")
                    LDIF.write(f"newrdn: {new_dn_val}\n")
                    LDIF.write("deleteoldrdn: 1\n")
                    LDIF.write("\n")
                    mrdnc -= 1
                    # Revise the DN list: add the new DN, and remove the old DN
                    entry_dn_list.append(new_dn_val)
                    if dn_val in entry_dn_list:
                        entry_dn_list.remove(dn_val)

                # Update the total count
                total_ops -= 1
        else:
            # Sequentially do each type of operation

            # Do the Adds
            addc = int(props['addUsers'])
            while addc != 0:
                dn = write_generic_user(
                    LDIF, addc, props['addUsers'], props['parent'],
                    name="addUser", changetype="\nchangetype: add")
                addc -= 1
                entry_dn_list.append(dn)

            # Mods
            while modc != 0:
                attr = randomPick(props['modAttrs'])
                val = (''.join((random.choice(RANDOM_CHARS) for i in range(0, random.randint(10, 30)))))
                LDIF.write(f"dn: uid=user{get_index(modc, props['numUsers'])},{props['parent']}\n")
                LDIF.write("changetype: modify\n")
                LDIF.write(f"replace: {attr}\n")
                LDIF.write(f"{attr}: {val}\n")
                LDIF.write("\n")
                modc -= 1

            # Modrdns
            while mrdnc != 0:
                dn_val = f"uid=user{get_index(mrdnc, props['numUsers'])},{props['parent']}"
                new_dn_val = f"cn=user{get_index(mrdnc, props['numUsers'])},{props['parent']}"
                LDIF.write(f"dn: {dn_val}\n")
                LDIF.write("changetype: modrdn\n")
                LDIF.write(f"newrdn: {new_dn_val}\n")
                LDIF.write("deleteoldrdn: 1\n")
                LDIF.write("\n")
                mrdnc -= 1
                # Revise the DN list: add the new DN, and remove the old DN
                entry_dn_list.append(new_dn_val)
                if dn_val in entry_dn_list:
                    entry_dn_list.remove(dn_val)

            # Deletes
            while delc != 0:
                dn_val = f"uid=user{get_index(delc, props['numUsers'])},{props['parent']}"
                LDIF.write(f"dn: {dn_val}\n")
                LDIF.write("changetype: delete\n")
                LDIF.write(" \n")
                delc -= 1
                if dn_val in entry_dn_list:
                    entry_dn_list.remove(dn_val)

        # Cleanup - delete all known entries
        if props['deleteUsers']:
            for dn in entry_dn_list:
                LDIF.write(f"dn: {dn}\n")
                LDIF.write("changetype: delete\n")
                LDIF.write(" \n")


def build_recursive_nodes(LDIF, dn, node_limit, max_entries):
    """
    Recursively create two nodes under each node, continue until there are no
    more max_entries.
    """
    global node_count
    wrote_node1 = False
    wrote_node2 = False

    # Create containers for DN1 and DN2
    dn1 = "ou=1," + dn
    LDIF.write(f'dn: {dn1}\n')
    LDIF.write('objectclass: top\n')
    LDIF.write('objectclass: organizationalUnit\n')
    LDIF.write('ou: ou=1\n\n')

    dn2 = "ou=2," + dn
    LDIF.write(f'dn: {dn2}\n')
    LDIF.write('objectclass: top\n')
    LDIF.write('objectclass: organizationalUnit\n')
    LDIF.write('ou: ou=2\n\n')

    # Add entries under each node
    for entry_idx in range(1, node_limit + 1):
        write_generic_user(LDIF, entry_idx, node_limit, dn1)
        max_entries -= 1
        if not wrote_node1:
            wrote_node1 = True
            node_count += 1
        if max_entries == 0:
            break

        write_generic_user(LDIF, entry_idx, node_limit, dn2)
        max_entries -= 1
        if not wrote_node2:
            wrote_node2 = True
            node_count += 1
        if max_entries == 0:
            break

    # Are we out of entries
    if max_entries == 0:
        return

    # Get the remaining entries to be split between two nodes
    new_node_max = max_entries // 2
    if (max_entries % 2) > 0:
        remainder = 1
    else:
        remainder = 0

    # Recursively build child nodes
    build_recursive_nodes(LDIF, dn1, node_limit, new_node_max)
    build_recursive_nodes(LDIF, dn2, node_limit, new_node_max + remainder)


def dbgen_nested_ldif(instance, ldif_file, props):
    """
    Create a deeply nested LDIF

        props = {
            "numUsers": ####   --> Total number of user entries to create
            'nodeLimit': ####  --> max number of entries to put into each node
            "suffix": DN
        }
    """

    global node_count
    node_count = 0
    with open(ldif_file, 'w') as LDIF:
        # Create the top suffix
        LDIF.write(get_node(props['suffix']))

        # Create all the nodes
        build_recursive_nodes(LDIF, props['suffix'], props['nodeLimit'], props['numUsers'])

    finalize_ldif_file(instance, ldif_file)

    return node_count



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
from datetime import datetime
from lib389.cli_base import CustomHelpFormatter
from lib389.tasks import BackupTask, RestoreTask

def backup_create(inst, basedn, log, args):
    log = log.getChild('backup_create')

    backup_task = BackupTask(inst)

    if args.archive is None:
        backup_dir_name = "backup-%s" % datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        archive = os.path.join(inst.ds_paths.backup_dir, backup_dir_name)
    else:
        archive = args.archive
    task_properties = {'nsArchiveDir': archive}

    if args.db_type is not None:
        task_properties['nsDatabaseType'] = args.db_type

    backup_task.create(properties=task_properties)
    if args.watch:
        log.info("Creating backup: %s", archive)
        backup_task.watch()
    else:
        backup_task.wait(timeout=args.timeout)

    result = backup_task.get_exit_code()
    if backup_task.is_complete() and result == 0:
        log.info("The backup create task has finished successfully")
    else:
        if result is None:
            raise ValueError("The backup create task has not completed. Please check server's error log for more information")
        else:
            raise ValueError(f"The backup create task has failed with the error code: ({result})")


def backup_restore(inst, basedn, log, args):
    log = log.getChild('backup_restore')

    restore_task = RestoreTask(inst)
    task_properties = {'nsArchiveDir': args.archive}
    if args.db_type is not None:
        task_properties['nsDatabaseType'] = args.db_type
    restore_task.create(properties=task_properties)
    if args.watch:
        log.info("Restoring backup: %s", args.archive)
        restore_task.watch()
    else:
        restore_task.wait(timeout=args.timeout)

    result = restore_task.get_exit_code()
    task_log = restore_task.get_task_log()

    if restore_task.is_complete() and result == 0:
        log.info("The backup restore task has finished successfully")
    else:
        if result is None:
            raise ValueError(f"The backup restore task has not completed. Please check server's error log for more information\n{task_log}")
        else:
            raise ValueError(f"The backup restore task has failed with the error code: {result}\n{task_log}")


def create_parser(subparsers):
    backup_parser = subparsers.add_parser('backup', help="Manage online backups", formatter_class=CustomHelpFormatter)

    subcommands = backup_parser.add_subparsers(help="action")

    create_backup_parser = subcommands.add_parser('create', help="Creates a backup of the database", formatter_class=CustomHelpFormatter)
    create_backup_parser.set_defaults(func=backup_create)
    create_backup_parser.add_argument('archive', nargs='?', default=None,
                                      help="Sets the directory where to store the backup files. "
                                           "Format: instance_name-year_month_date_hour_minutes_seconds. "
                                           "Default: /var/lib/dirsrv/slapd-instance/bak/ ")
    create_backup_parser.add_argument('-t', '--db-type', default="ldbm database",
                                      help="Sets the database type. Default: ldbm database")
    create_backup_parser.add_argument('--timeout', type=int, default=120,
                                      help="Sets the task timeout.  Default is 120 seconds,")
    create_backup_parser.add_argument('--watch', action='store_true', help='Watch the status of the backup task')

    restore_parser = subcommands.add_parser('restore', help="Restores a database from a backup", formatter_class=CustomHelpFormatter)
    restore_parser.set_defaults(func=backup_restore)
    restore_parser.add_argument('archive', help="Set the directory that contains the backup files")
    restore_parser.add_argument('-t', '--db-type', default="ldbm database",
                                help="Sets the database type. Default: ldbm database")
    restore_parser.add_argument('--timeout', type=int, default=120,
                                help="Sets the task timeout.  Default is 120 seconds.")
    restore_parser.add_argument('--watch', action='store_true', help='Watch the status of the restore task')



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 William Brown <william@blackhats.net.au>
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---


from lib389.plugins import EntryUUIDPlugin, EntryUUIDFixupTasks
from lib389.cli_conf import add_generic_plugin_parsers
from lib389.utils import get_task_status
from lib389.cli_base import CustomHelpFormatter


def do_fixup(inst, basedn, log, args):
    plugin = EntryUUIDPlugin(inst)
    log.info('Attempting to add task entry...')
    if not plugin.status():
        log.error("'%s' is disabled. Fix up task can't be executed" % plugin.rdn)
        return
    fixup_task = plugin.fixup(args.DN, args.filter)
    if args.wait:
        log.info(f'Waiting for fixup task "{fixup_task.dn}" to complete.  You can safely exit by pressing Control C ...')
        fixup_task.wait(timeout=args.timeout)
        exitcode = fixup_task.get_exit_code()
        if exitcode != 0:
            if exitcode is None:
                raise ValueError(f'EntryUUID fixup task "{fixup_task.dn}" for {args.DN} has not completed. Please, check logs')
            else:
                raise ValueError(f'EntryUUID fixup task "{fixup_task.dn}" for {args.DN} has failed (error {exitcode}). Please, check logs')
        else:
            log.info('Fixup task successfully completed')
    else:
        log.info(f'Successfully added task entry "{fixup_task.dn}". This task is running in the background. To track its progress you can use the "fixup-status" command.')


def do_fixup_status(inst, basedn, log, args):
    get_task_status(inst, log, EntryUUIDFixupTasks, dn=args.dn, show_log=args.show_log,
                    watch=args.watch, use_json=args.json)


def create_parser(subparsers):
    referint = subparsers.add_parser('entryuuid', help='Manage and configure EntryUUID plugin', formatter_class=CustomHelpFormatter)
    subcommands = referint.add_subparsers(help='action')

    add_generic_plugin_parsers(subcommands, EntryUUIDPlugin)

    fixup = subcommands.add_parser('fixup', help='Run the fix-up task for EntryUUID plugin', formatter_class=CustomHelpFormatter)
    fixup.set_defaults(func=do_fixup)
    fixup.add_argument('DN', help="Base DN that contains entries to fix up")
    fixup.add_argument('-f', '--filter',
                       help='Filter for entries to fix up.\n If omitted, all entries under base DN'
                            'will have their EntryUUID attribute regenerated if not present.')
    fixup.add_argument('--wait', action='store_true',
                       help="Wait for the task to finish, this could take a long time")
    fixup.add_argument('--timeout', type=int, default=0,
                       help="Sets the task timeout. Default is 0 (no timeout)")

    fixup_status = subcommands.add_parser('fixup-status', help='Check the status of a fix-up task', formatter_class=CustomHelpFormatter)
    fixup_status.set_defaults(func=do_fixup_status)
    fixup_status.add_argument('--dn', help="The task entry's DN")
    fixup_status.add_argument('--show-log', action='store_true', help="Display the task log")
    fixup_status.add_argument('--watch', action='store_true',
                       help="Watch the task's status and wait for it to finish")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json

from lib389.pwpolicy import PwPolicyManager
from lib389.idm.user import (
    nsUserAccount,
    nsUserAccounts,
    BasicUserAccount,
    BasicUserAccounts,
    TraditionalUserAccount,
    TraditionalUserAccounts,
)
from lib389.idm.services import ServiceAccount, ServiceAccounts
from lib389.cli_base import populate_attr_arguments, _generic_modify, CustomHelpFormatter
from lib389.cli_idm import (
    _generic_list,
    _generic_get,
    _generic_get_dn,
    _generic_create,
    _generic_rename,
    _generic_delete,
    _get_arg,
    _get_attributes,
    _warn,
    )

DEFAULT_USER_TYPE = "posix"
USER_TYPE_CHOICES = ['basic', 'posix', 'service', 'traditional']
SINGULAR_DICT = {
    None: nsUserAccount,
    'posix': nsUserAccount,
    'basic': BasicUserAccount,
    'service': ServiceAccount,
    'traditional': TraditionalUserAccount,
}
MANY_DICT = {
    None: nsUserAccounts,
    'posix': nsUserAccounts,
    'basic': BasicUserAccounts,
    'service': ServiceAccounts,
    'traditional': TraditionalUserAccounts,
}

RDN = 'uid'


# These are a generic specification, try not to tamper with them

def list(inst, basedn, log, args):
    _generic_list(inst, basedn, log.getChild('_generic_list'), MANY_DICT[args.user_type], args)


def get(inst, basedn, log, args):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_get(inst, basedn, log.getChild('_generic_get'), MANY_DICT[args.user_type], rdn, args)


def get_dn(inst, basedn, log, args):
    dn = _get_arg( args.dn, msg="Enter dn to retrieve")
    _generic_get_dn(inst, basedn, log.getChild('_generic_get_dn'), MANY_DICT[args.user_type], dn, args)


def create(inst, basedn, log, args):
    user_type = getattr(args, 'user_type', None) or DEFAULT_USER_TYPE
    kwargs = _get_attributes(args, SINGULAR_DICT[user_type]._must_attributes)
    _generic_create(inst, basedn, log.getChild('_generic_create'), MANY_DICT[args.user_type], kwargs, args)


def delete(inst, basedn, log, args, warn=True):
    dn = _get_arg( args.dn, msg="Enter dn to delete")
    if warn:
        _warn(dn, msg="Deleting %s %s" % (SINGULAR_DICT[args.user_type].__name__, dn))
    _generic_delete(inst, basedn, log.getChild('_generic_delete'), SINGULAR_DICT[args.user_type], dn, args)


def modify(inst, basedn, log, args, warn=True):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_modify(inst, basedn, log.getChild('_generic_modify'), MANY_DICT[args.user_type], rdn, args)


def rename(inst, basedn, log, args, warn=True):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_rename(inst, basedn, log.getChild('_generic_rename'), MANY_DICT[args.user_type], rdn, args)


def get_pwp(inst, basedn, log, args):
    """Show the effective password policy for a user account."""
    log = log.getChild('get_pwp')
    selector = _get_arg(args.selector, msg="Enter username to look up password policy for")
    user_type = getattr(args, 'user_type', None) or DEFAULT_USER_TYPE
    pwp_manager = PwPolicyManager(inst)

    report = pwp_manager.get_effective_policy(basedn, user_type, selector, parent_basedn=args.parent_dn)
    if args.json:
        log.info(json.dumps(pwp_manager.format_report_json(report), indent=4))
    else:
        log.info(pwp_manager.format_report_text(report))


def create_parser(subparsers, user_type=DEFAULT_USER_TYPE):
    user_parser = subparsers.add_parser('user',
                                        help='Manage posix users.  The organizationalUnit (by default "ou=people") needs '
                                             'to exist prior to managing users.')
    subcommands = user_parser.add_subparsers(help='action')
    user_parser.add_argument('--user-type',
                             choices=USER_TYPE_CHOICES,
                             default=DEFAULT_USER_TYPE,
                             help=f'''The type of user to manage.
                                  "basic" is an entry based off of objectclasses:
                                      nsPerson, nsAccount, and nsOrgPerson.
                                  "posix" is based off of objectclasses:
                                      nsPerson, nsAccount, nsOrgPerson, and posixAccount.
                                  "service" is based off of objectclasses:
                                      nsAccount and applicationProcess.
                                  "traditional", aka legacy, is based off of objectclasses:
                                      person, organizationalPerson, and inetOrgPerson.
                                  The default user type is "{DEFAULT_USER_TYPE}"''')

    list_parser = subcommands.add_parser('list', help='list', formatter_class=CustomHelpFormatter)
    list_parser.set_defaults(func=list)
    list_parser.add_argument('--full-dn', action='store_true',
                             help="Return the full DN of the entry instead of the RDN value")

    get_parser = subcommands.add_parser('get', help='get', formatter_class=CustomHelpFormatter)
    get_parser.set_defaults(func=get)
    get_parser.add_argument('selector', nargs='?', help='The term to search for')

    get_pwp_parser = subcommands.add_parser(
        'get-pwp',
        help='Show the effective password policy for a user',
        formatter_class=CustomHelpFormatter,
        description='Resolve and display the password policy (global, user local, or subtree '
                    'local) that applies to the user, including all effective settings.',
    )
    get_pwp_parser.set_defaults(func=get_pwp)
    get_pwp_parser.add_argument(
        'selector',
        nargs='?',
        help='The user identifier to look up (same as user get, typically uid)',
    )
    get_pwp_parser.add_argument(
        '--parent-dn',
        default=None,
        help='A base DN to use for user lookups. Default is "ou=people,$SUFFIX" '
             'for users, or "ou=services,$SUFFIX" for service accounts',
    )

    get_dn_parser = subcommands.add_parser('get_dn', help='get_dn', formatter_class=CustomHelpFormatter)
    get_dn_parser.set_defaults(func=get_dn)
    get_dn_parser.add_argument('dn', nargs='?', help='The dn to get')

    create_user_parser = subcommands.add_parser('create', help='create', formatter_class=CustomHelpFormatter)
    create_user_parser.set_defaults(func=create)
    populate_attr_arguments(create_user_parser, SINGULAR_DICT[user_type]._must_attributes)

    modify_parser = subcommands.add_parser('modify', help='modify <add|delete|replace>:<attribute>:<value> ...', formatter_class=CustomHelpFormatter)
    modify_parser.set_defaults(func=modify)
    modify_parser.add_argument('selector', nargs=1, help='The %s to modify' % RDN)
    modify_parser.add_argument('changes', nargs='+', help="A list of changes to apply in format: <add|delete|replace>:<attribute>:<value>")

    rename_parser = subcommands.add_parser('rename', help='rename the object', formatter_class=CustomHelpFormatter)
    rename_parser.set_defaults(func=rename)
    rename_parser.add_argument('selector', help='The %s to modify' % RDN)
    rename_parser.add_argument('new_name', help='A new user name')
    rename_parser.add_argument('--keep-old-rdn', action='store_true', help="Specify whether the old RDN (i.e. 'cn: old_user') should be kept as an attribute of the entry or not")

    delete_parser = subcommands.add_parser('delete', help='deletes the object', formatter_class=CustomHelpFormatter)
    delete_parser.set_defaults(func=delete)
    delete_parser.add_argument('dn', nargs='?', help='The dn to delete')

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

import ldap
import logging

from lib389.referral import Referrals, Referral

from lib389.tests.topologies import topology_st

from lib389._constants import DEFAULT_SUFFIX

log = logging.getLogger(__name__)

def test_referral(topology_st):
    standalone = topology_st.standalone

    rs = Referrals(standalone, DEFAULT_SUFFIX)

    r = rs.create(properties={
        'cn': 'testref',
        'ref': 'ldap://localhost:38901/ou=People,dc=example,dc=com'
    })

    r_all = rs.list()
    assert(len(r_all) == 1)
    r2 = r_all[0]
    assert(r2.present('ref', 'ldap://localhost:38901/ou=People,dc=example,dc=com'))


